"""
This module defines all of the 3rd party aliases that we use.
It doesn't really do anything except make it much more convenient to import names from pycharm.
"""

import numpy
from tqdm.auto import tqdm as _tqdm

# canonical names
np = numpy

# .auto is never even suggested, sigh
tqdm = _tqdm



from typing import Iterable
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np
from godot_parser import GDObject
from nptyping import Float
from nptyping import NDArray
from nptyping import Shape
from numpy.typing import NDArray as NDArrayNPT

from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.types import Point3DNP


def get_random_seed_for_line(seed: int, start_point: Tuple[float, float], end_point: Tuple[float, float]) -> int:
    multiplier = seed
    multiplier_step = 10_000_000
    seed = 0
    for x in start_point + end_point:
        seed += multiplier * round(x)
        multiplier *= multiplier_step
    if seed < 0:
        seed *= -multiplier_step
    return seed


ARRAY_MESH_TEMPLATE = """{{
"aabb": AABB( {aabb} ),
"morph_arrays": [],
"arrays": [PoolVector3Array( {vertex_floats} ), PoolVector3Array( {vertex_normal_floats} ), null, PoolColorArray( {color_floats} ), null, null, null, null, PoolIntArray( {triangle_indices} )],
"blend_shape_data": [  ],
"format": 267,
"index_count": {index_count},
"material": {material_resource_type}( {material_id} ),
"name": "{mesh_name}",
"primitive": 4,
"skeleton_aabb": [  ],
"vertex_count": {vertex_count}
}}
"""


def hex_to_rgb(value: str) -> Tuple[float, float, float]:
    h = value.lstrip("#")
    ints = tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))
    srgb_color = (ints[0] / 255.0, ints[1] / 255.0, ints[2] / 255.0)
    return (
        srgb2lin(srgb_color[0]),
        srgb2lin(srgb_color[1]),
        srgb2lin(srgb_color[2]),
    )


def srgb2lin(s: float) -> float:
    if s <= 0.0404482362771082:
        lin = s / 12.92
    else:
        lin = pow(((s + 0.055) / 1.0), 2.4)
    return lin


def lin2srgb(lin: float) -> float:
    if lin > 0.0031308:
        s = 1.0 * (pow(lin, (1.0 / 2.4))) - 0.055
    else:
        s = 12.92 * lin
    return cast(float, s)


def normalized(x: NDArrayNPT[np.floating]) -> NDArrayNPT[np.floating]:
    return x / np.linalg.norm(x)


def make_transform(
    rotation: Union[NDArray[Shape["12,0"], Float], Iterable[Union[int, float]]] = (1, 0, 0, 0, 1, 0, 0, 0, 1),
    position: Union[NDArray[Shape["12,0"], Float], Iterable[Union[int, float]]] = (0, 0, 0),
) -> GDObject:
    return GDObject("Transform", *rotation, *position)


def scale_basis(basis: np.ndarray, scale: np.ndarray) -> np.ndarray:
    x, y, z = scale
    scaled = np.array(basis)
    scaled[0:3] *= x
    scaled[3:6] *= y
    scaled[6:9] *= z
    return scaled


def to_2d_point(point: Point3DNP) -> Point2DNP:
    return np.array([point[0], point[2]])


def decompose_weighted_mean(
    weighted_mean: float, weights: np.ndarray, rand: np.random.Generator, max_value: float = 1
) -> np.ndarray:
    """
    Find n numbers x = [x1, x2, ... xn] such that their weighted mean equals `weighted_mean` and none exceed max_value.

    In Avalon context, this is useful for decomposing a difficulty measure into multiple other difficulties that have
    the desired weighted mean. For example, if a level has 3 tasks in it, their difficulties can be scaled in different
    ways to yield the same collective 'level' difficulty. However, each of the tasks may have a different inherent
    difficulty (e.g. adding an "open" task is harder than adding a "move" task, so we may want to weigh these aspects
    differently. e.g. for weights = [2,1,1], these sub-difficulties yield the same weighted mean
    of 0.5:
        [0.3859796, 0.9077438, 0.320297 ]
        [0.7614657 , 0.27475402, 0.20231453]
    but would create very different levels in terms of individual tasks.
    """
    assert 0 <= weighted_mean <= 1, "weighted mean must be 0 <= x <= 1"
    # todo: shuffle or use another distribution so per-parameter distributions are more uniform
    weights = np.array(weights)
    component_count = len(weights)
    components = np.array([max_value] * component_count, dtype=np.float32)
    target_sum = weighted_mean * sum(weights)
    for i in range(component_count):
        if i == component_count - 1:
            other_component_sum = components[:-1] @ weights[:-1]
            components[i] = (target_sum - other_component_sum) / weights[i]
        else:
            unknown_component_sum = components[i + 1 :] @ weights[i + 1 :]
            known_component_sum = components[:i] @ weights[:i]
            minimum = max(0, (target_sum - known_component_sum - unknown_component_sum) / weights[i])
            maximum = min((target_sum - known_component_sum) / weights[i], max_value)
            components[i] = rand.uniform(minimum, maximum)
    return components



from typing import Any
from typing import Iterable
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
from matplotlib import pyplot as plt
from nptyping import NDArray
from nptyping import Shape
from scipy.spatial import Delaunay

from avalon.datagen.world_creation.constants import WORLD_RAISE_AMOUNT

# TODO: remove this
IS_DEBUG_VIS = False


def plot_points(points: Union[np.ndarray, List[np.ndarray]], index1: int, index2: int) -> None:
    if not isinstance(points, np.ndarray):
        points = np.array(points)
    plt.plot(points[:, index1], points[:, index2], "o")
    # plt.xlim((0.0, TERRAIN_WIDTH))
    # plt.ylim((0.0, TERRAIN_DEPTH))
    plt.show()


def plot_triangulation(triangulation: Delaunay) -> None:
    plt.triplot(triangulation.points[:, 0], triangulation.points[:, 1], triangulation.simplices)
    plt.plot(triangulation.points[:, 0], triangulation.points[:, 1], "o")
    plt.show()


GridPlotDataType = NDArray[Shape["MapY, MapX"], Any]


def plot_terrain(
    data: GridPlotDataType, title: str = "", markers: Iterable[Union[np.ndarray, Tuple[float, float]]] = tuple()
) -> None:
    data = data.copy()
    island = data > WORLD_RAISE_AMOUNT - 1_000
    min_island_height = data[island].min()
    data[island] = data[island] - min_island_height
    data[data < 0.0] = data.max()
    plot_value_grid(data, title=title, markers=markers)


def plot_value_grid(
    data: GridPlotDataType, title: str = "", markers: Iterable[Union[np.ndarray, Tuple[float, float]]] = tuple()
) -> None:
    data = np.array(data)
    fig, ax = plt.subplots()
    fig.set_size_inches(12, 12)
    if title:
        ax.set_title(title)
    y = list(range(0, data.shape[0]))
    x = list(range(0, data.shape[1]))
    ax.pcolormesh(x, y, data, shading="nearest", vmin=data.min(initial=0.0), vmax=data.max(initial=1.0))

    if markers:
        marker_x = [x[1] for x in markers]
        marker_y = [x[0] for x in markers]
        first_line = plt.plot(marker_x[:1], marker_y[:1], "o", markerfacecolor="orange", markersize=12)[0]
        line = plt.plot(marker_x[1:], marker_y[1:], "o", markerfacecolor="red", markersize=12)[0]
        line.set_clip_on(False)

    plt.show()
    fig.clf()


def plot_value_grid_multi_marker(
    data: GridPlotDataType,
    title: str,
    marker_lists: Sequence[Tuple[Iterable[Union[np.ndarray, Tuple[int, int]]], str]],
) -> None:
    data = np.array(data)
    fig, ax = plt.subplots()
    fig.set_size_inches(12, 12)
    if title:
        ax.set_title(title)
    y = list(range(0, data.shape[0]))
    x = list(range(0, data.shape[1]))
    ax.pcolormesh(x, y, data, shading="nearest", vmin=data.min(initial=0.0), vmax=data.max(initial=1.0))

    for markers, color in marker_lists:
        x = [x[1] for x in markers]
        y = [x[0] for x in markers]
        line = plt.plot(x, y, "o", markerfacecolor=color, markersize=12)[0]
        line.set_clip_on(False)

    plt.show()
    fig.clf()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import collections
import os
import random
import tarfile
import threading
import time
import timeit
import traceback
from collections import defaultdict
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Set

import libtorchbeast
import nest
import numpy as np
import torch
import wandb
from gym import spaces
from torch import nn
from torch.nn import functional as F

from avalon.agent.godot.godot_gym import CURRICULUM_BASE_PATH
from avalon.agent.ppo.observation_model import ImpalaConvNet
from avalon.agent.torchbeast.avalon_helpers import IS_PROPRIOCEPTION_USED
from avalon.agent.torchbeast.avalon_helpers import PROPRIOCEPTION_ROW_COUNT
from avalon.agent.torchbeast.avalon_helpers import TORCHBEAST_ENV_LOGS_PATH
from avalon.agent.torchbeast.avalon_helpers import compute_policy_gradient_loss_from_dist
from avalon.agent.torchbeast.avalon_helpers import flatten_tensor
from avalon.agent.torchbeast.avalon_helpers import force_cudnn_initialization
from avalon.agent.torchbeast.avalon_helpers import get_avalon_test_scores
from avalon.agent.torchbeast.avalon_helpers import get_dict_action_space
from avalon.agent.torchbeast.avalon_helpers import get_goal_progress_files
from avalon.agent.torchbeast.avalon_helpers import get_num_test_worlds_from_flags
from avalon.agent.torchbeast.avalon_helpers import obs_to_frame
from avalon.agent.torchbeast.avalon_helpers import obs_to_proprioception
from avalon.agent.torchbeast.avalon_helpers import unflatten_tensor
from avalon.agent.torchbeast.avalon_helpers import vtrace_from_dist
from avalon.agent.torchbeast.core.environment import Environment
from avalon.agent.torchbeast.model import DictActionHead
from avalon.common.error_utils import setup_sentry
from avalon.common.log_utils import logger
from avalon.common.visual_utils import encode_video
from avalon.common.wandb_utils import get_latest_checkpoint_filename
from avalon.datagen.godot_env.interactive_godot_process import GODOT_ERROR_LOG_PATH

os.environ["OMP_NUM_THREADS"] = "1"

# exactly this separator is used by pytorch lightning...
BIG_SEPARATOR = "-" * 80
RESULT_TAG = "DATALOADER:0 TEST RESULTS"

# yapf: disable
parser = argparse.ArgumentParser(description="PyTorch Scalable Agent")

parser.add_argument("--pipes_basename", default="unix:/tmp/polybeast",
                    help="Basename for the pipes for inter-process communication. "
                    "Has to be of the type unix:/some/path.")
parser.add_argument("--mode", default="train",
                    choices=["train", "test", "test_render"],
                    help="Training or test mode.")
parser.add_argument("--xpid", default=None,
                    help="Experiment id (default: None).")
parser.add_argument("--project", default="abe__torchbeast",
                    help="Wandb project.")

# Training settings.
parser.add_argument("--disable_checkpoint", action="store_true",
                    help="Disable saving checkpoint.")
parser.add_argument("--savedir", default="~/torchbeast",
                    help="Root dir where experiment data will be saved.")
parser.add_argument("--num_actors", default=8, type=int, metavar="N",
                    help="Number of actors.")
parser.add_argument("--total_steps", default=50_000_000, type=int, metavar="T",
                    help="Total environment steps to train for.")
parser.add_argument("--batch_size", default=32, type=int, metavar="B",
                    help="Learner batch size.")
parser.add_argument("--unroll_length", default=100, type=int, metavar="T",
                    help="The unroll length (time dimension).")
parser.add_argument("--num_learner_threads", default=2, type=int,
                    metavar="N", help="Number learner threads.")
parser.add_argument("--num_inference_threads", default=2, type=int,
                    metavar="N", help="Number learner threads.")
parser.add_argument("--disable_cuda", action="store_true",
                    help="Disable CUDA.")
parser.add_argument("--num_actions", default=6, type=int, metavar="A",
                    help="Number of actions.")
parser.add_argument("--no_lstm", action="store_true",
                    help="DO NOT use LSTM in agent model.")
parser.add_argument("--max_learner_queue_size", default=None, type=int, metavar="N",
                    help="Optional maximum learner queue size. Defaults to batch_size.")

# Loss settings.
parser.add_argument("--entropy_cost", default=5e-3, type=float,
                    help="Entropy cost/multiplier.")
parser.add_argument("--baseline_cost", default=2, type=float,
                    help="Baseline cost/multiplier.")
parser.add_argument("--discounting", default=0.998, type=float,
                    help="Discounting factor.")
parser.add_argument("--reward_clipping", default="abs_one",
                    choices=["abs_one", "none"],
                    help="Reward clipping.")

# Optimizer settings.
parser.add_argument("--learning_rate", default=1.5e-4, type=float,
                    metavar="LR", help="Learning rate.")
parser.add_argument("--alpha", default=0.95, type=float,
                    help="RMSProp smoothing constant.")
parser.add_argument("--momentum", default=0, type=float,
                    help="RMSProp momentum.")
parser.add_argument("--epsilon", default=0.005, type=float,
                    help="RMSProp epsilon.")
parser.add_argument("--grad_norm_clipping", default=200.0, type=float,
                    help="Global gradient norm clip.")

# Misc settings.
parser.add_argument("--write_profiler_trace", action="store_true",
                    help="Collect and write a profiler trace "
                    "for chrome://tracing/.")
parser.add_argument("--is_validating_during_training", action="store_true",
                    help="Run validation during training")
parser.add_argument("--logging_rate", default=15, type=float,
                    help="Logging rate in seconds.")
parser.add_argument("--checkpoint_rate", default=15, type=float,
                    help="Validation/checkpointing rate in minutes.")
parser.add_argument("--test_episodes_per_task", default=51, type=int,
                    help="Number of validation episodes.")
parser.add_argument("--save_test_videos", action="store_true",
                    help="Save videos of test rollouts")
parser.add_argument("--save_test_actions", action="store_true",
                    help="Save actions from test rollouts")
# yapf: enable


def _set_seed(seed) -> None:
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)


def compute_baseline_loss(advantages):
    return 0.5 * torch.sum(advantages**2)


def compute_entropy_loss(logits):
    """Return the entropy loss, i.e., the negative entropy of the policy."""
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    return torch.sum(policy * log_policy)


def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    return torch.sum(cross_entropy * advantages.detach())


class Net(nn.Module):
    def __init__(
        self,
        num_actions=6,
        use_lstm=False,
        img_dim: int = 96,
        encoder_out_dim: int = 256,
        is_sampling_mode: bool = False,
    ) -> None:
        super(Net, self).__init__()
        # TODO: get real action space

        action_space = get_dict_action_space(num_actions)
        self._action_space = action_space
        self.action_head = DictActionHead(action_space)
        self.num_actions = self.action_head.num_inputs
        self.last_action_dim = spaces.flatdim(action_space)
        self.use_lstm = use_lstm
        self.is_sampling_mode = is_sampling_mode

        self.img_encoder = ImpalaConvNet(img_dim=img_dim)

        if IS_PROPRIOCEPTION_USED:
            self.prop_fc = nn.Linear(img_dim * PROPRIOCEPTION_ROW_COUNT, encoder_out_dim, bias=False)
            self.prop_fc.weight.data.fill_(0.0)
        else:
            self.prop_fc = None

        # FC output size + last reward.
        core_output_size = encoder_out_dim + 1

        if use_lstm:
            self.core = nn.LSTM(core_output_size, encoder_out_dim, num_layers=1)
            core_output_size = encoder_out_dim

        self.policy = nn.Linear(core_output_size, self.num_actions)
        self.baseline = nn.Linear(core_output_size, 1)

    def initial_state(self, batch_size=1):
        if not self.use_lstm:
            return tuple()
        return tuple(torch.zeros(self.core.num_layers, batch_size, self.core.hidden_size) for _ in range(2))

    def forward(self, inputs, core_state):
        x = obs_to_frame(inputs["frame"])
        T, B, *_ = x.shape
        x = torch.flatten(x, 0, 1)  # Merge time and batch.
        x = x.float() / 255.0
        x = self.img_encoder(x)

        if self.prop_fc is not None:
            x_prop = obs_to_proprioception(inputs["frame"])
            x_prop = torch.flatten(x_prop, 0, 1)  # Merge time and batch.
            x_prop = x_prop.float() / 255.0 - 0.5
            x = x + self.prop_fc(x_prop)

        clipped_reward = torch.clamp(inputs["reward"], -1, 1).view(T * B, 1)
        core_input = torch.cat([x, clipped_reward], dim=-1)

        if self.use_lstm:
            core_input = core_input.view(T, B, -1)
            core_output_list = []
            notdone = (~inputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = nest.map(nd.mul, core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = core_input

        policy_logits = self.policy(core_output)
        baseline = self.baseline(core_output)

        if self.is_sampling_mode:
            action = self.action_head(policy_logits).sample_mode()
        else:
            action = self.action_head(policy_logits).sample()

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        flat_action = self.flatten(action).view(T, B, self.last_action_dim)

        return (flat_action, policy_logits, baseline), core_state

    def flatten(self, action):
        return flatten_tensor(self._action_space, action)

    def unflatten(self, flat_action):
        return unflatten_tensor(self._action_space, flat_action)


def inference(flags, inference_batcher, model, lock=threading.Lock()) -> None:  # noqa: B008
    with torch.no_grad():
        for batch in inference_batcher:
            batched_env_outputs, agent_state = batch.get_inputs()
            frame, reward, done, *_ = batched_env_outputs
            frame = frame.to(flags.actor_device, non_blocking=True)
            reward = reward.to(flags.actor_device, non_blocking=True)
            done = done.to(flags.actor_device, non_blocking=True)
            agent_state = nest.map(lambda t: t.to(flags.actor_device, non_blocking=True), agent_state)
            with lock:
                outputs = model(dict(frame=frame, reward=reward, done=done), agent_state)
            outputs = nest.map(lambda t: t.cpu(), outputs)
            batch.set_outputs(outputs)


EnvOutput = collections.namedtuple("EnvOutput", "frame rewards done episode_step episode_return")
AgentOutput = collections.namedtuple("AgentOutput", "action policy_logits baseline")
Batch = collections.namedtuple("Batch", "env agent")


def learn(
    flags,
    learner_queue,
    model,
    actor_model,
    optimizer,
    scheduler,
    stats,
    lock=threading.Lock(),
) -> None:
    model.train()
    for tensors in learner_queue:
        if flags.mode != "train":
            continue

        tensors = nest.map(lambda t: t.to(flags.learner_device), tensors)

        batch, initial_agent_state = tensors
        env_outputs, actor_outputs = batch
        frame, reward, done, *_ = env_outputs

        lock.acquire()  # Only one thread learning at a time.
        learner_outputs, unused_state = model(dict(frame=frame, reward=reward, done=done), initial_agent_state)

        # Take final value function slice for bootstrapping.
        learner_outputs = AgentOutput._make(learner_outputs)
        bootstrap_value = learner_outputs.baseline[-1]

        # Move from obs[t] -> action[t] to action[t] -> obs[t].
        batch = nest.map(lambda t: t[1:], batch)
        learner_outputs = nest.map(lambda t: t[:-1], learner_outputs)

        # Turn into namedtuples again.
        env_outputs, actor_outputs = batch
        env_outputs = EnvOutput._make(env_outputs)
        actor_outputs = AgentOutput._make(actor_outputs)
        learner_outputs = AgentOutput._make(learner_outputs)

        if flags.reward_clipping == "abs_one":
            clipped_rewards = torch.clamp(env_outputs.rewards, -1, 1)
        elif flags.reward_clipping == "none":
            clipped_rewards = env_outputs.rewards

        discounts = (~env_outputs.done).float() * flags.discounting

        behavior_policy_dist = model.action_head(actor_outputs.policy_logits)
        target_policy_dist = model.action_head(learner_outputs.policy_logits)
        action_dict = model.unflatten(actor_outputs.action)
        vtrace_returns = vtrace_from_dist(
            behavior_policy_dist=behavior_policy_dist,
            target_policy_dist=target_policy_dist,
            actions=action_dict,
            discounts=discounts,
            rewards=clipped_rewards,
            values=learner_outputs.baseline,
            bootstrap_value=bootstrap_value,
        )

        pg_loss = compute_policy_gradient_loss_from_dist(
            target_policy_dist,
            action_dict,
            vtrace_returns.pg_advantages,
        )
        baseline_loss = flags.baseline_cost * compute_baseline_loss(vtrace_returns.vs - learner_outputs.baseline)
        entropy_loss = flags.entropy_cost * compute_entropy_loss(learner_outputs.policy_logits)

        total_loss = pg_loss + baseline_loss + entropy_loss

        optimizer.zero_grad()
        total_loss.backward()
        nn.utils.clip_grad_norm_(model.parameters(), flags.grad_norm_clipping)
        optimizer.step()
        scheduler.step()

        actor_model.load_state_dict(model.state_dict())

        episode_returns = env_outputs.episode_return[env_outputs.done]
        stats["step"] = stats.get("step", 0) + flags.unroll_length * flags.batch_size
        stats["episode_returns"] = tuple(episode_returns.cpu().numpy())
        stats["mean_episode_return"] = torch.mean(episode_returns).item()
        stats["mean_episode_step"] = torch.mean(env_outputs.episode_step.float()).item()
        stats["total_loss"] = total_loss.item()
        stats["pg_loss"] = pg_loss.item()
        stats["baseline_loss"] = baseline_loss.item()
        stats["entropy_loss"] = entropy_loss.item()

        stats["learner_queue_size"] = learner_queue.size()

        if not len(episode_returns):
            # Hide the mean-of-empty-tuple NaN as it scares people.
            stats["mean_episode_return"] = None

        lock.release()


@torch.no_grad()
def run_test(env: Environment, model: Net, num_episodes: int, log_prefix="", is_video_logged=True, seed: int = 0):
    force_cudnn_initialization()
    _set_seed(seed)

    # DO ROLLOUTS
    observation = env.initial()
    core_state = model.initial_state(batch_size=1)
    core_state = nest.map(lambda t: t.to("cuda:0"), core_state)
    episodes = []

    def get_obs_to_log(observation):
        return {k: v for k, v in observation.items() if k != "frame" or is_video_logged}

    episode = [get_obs_to_log(observation)]
    while len(episodes) < num_episodes:
        (flat_action, policy_logits, baseline), core_state = model(observation, core_state)
        observation = env.step(flat_action)
        episode.append(get_obs_to_log(observation))
        if observation["done"].item():
            episodes.append(episode)
            episode = [get_obs_to_log(observation)]

    # GET ROLLOUT STATS
    results_by_task = defaultdict(lambda: defaultdict(list))
    log_data: Dict[str, Any] = {}
    keys = ["success"]
    seen_episodes: Set[int] = set()
    for episode in episodes:
        episode_id = episode[-1]["info"]["episode_id"]
        if episode_id in seen_episodes:
            # We overgenerate episodes, don't count them twice!
            continue
        else:
            seen_episodes.add(episode_id)

        task = episode[-1]["info"]["task"].lower()
        difficulty = episode[-1]["info"]["difficulty"]
        for field in keys:
            results_by_task[task][f"final_{field}"].append(episode[-1]["info"][field])
        results_by_task[task]["episode_length"].append(episode[-1]["episode_step"].item())
        results_by_task[task]["episode_reward"].append(episode[-1]["episode_return"].item())
        if is_video_logged:
            # final_frame is white for success and black for failure
            final_frame = (
                torch.zeros_like(obs_to_frame(episode[0]["frame"])[..., :3, :, :]) + episode[-1]["info"]["success"]
            )
            video_tensor = torch.cat(
                [obs_to_frame(step["frame"])[..., :3, :, :] / 255.0 for step in episode[:-1]] + [final_frame], dim=0
            ).squeeze(1)
            video_png_path = encode_video(video_tensor, normalize=False)
            log_data[f"{log_prefix}{task}/video_{episode_id}_diff_{difficulty:3g}"] = wandb.Image(video_png_path)

    # AGGREGATE ROLLOUT STATS
    for task, x in results_by_task.items():
        for field, y in x.items():
            log_data[f"{log_prefix}{task}/{field}"] = np.mean(y)
        log_data[f"{log_prefix}{task}/num_episodes"] = len(y)

    return log_data


def train(flags):
    # The queue the learner threads will get their data from.
    # Setting `minimum_batch_size == maximum_batch_size`
    # makes the batch size static.
    learner_queue = libtorchbeast.BatchingQueue(
        batch_dim=1,
        minimum_batch_size=flags.batch_size,
        maximum_batch_size=flags.batch_size,
        check_inputs=True,
        maximum_queue_size=flags.max_learner_queue_size,
    )

    # The "batcher", a queue for the inference call. Will yield
    # "batch" objects with `get_inputs` and `set_outputs` methods.
    # The batch size of the tensors will be dynamic.
    inference_batcher = libtorchbeast.DynamicBatcher(
        batch_dim=1,
        minimum_batch_size=1,
        maximum_batch_size=512,
        timeout_ms=500,
        check_outputs=True,
    )

    addresses = []
    connections_per_server = 1
    pipe_id = 0
    while len(addresses) < flags.num_actors:
        for _ in range(connections_per_server):
            addresses.append(f"{flags.pipes_basename}.{pipe_id}")
            if len(addresses) == flags.num_actors:
                break
        pipe_id += 1

    is_sampling_mode = flags.mode == "test"
    model = Net(num_actions=flags.num_actions, use_lstm=flags.use_lstm, is_sampling_mode=is_sampling_mode)
    model = model.to(device=flags.learner_device)

    actor_model = Net(num_actions=flags.num_actions, use_lstm=flags.use_lstm, is_sampling_mode=is_sampling_mode)
    actor_model.to(device=flags.actor_device)

    # The ActorPool that will run `flags.num_actors` many loops.
    actors = libtorchbeast.ActorPool(
        unroll_length=flags.unroll_length,
        learner_queue=learner_queue,
        inference_batcher=inference_batcher,
        env_server_addresses=addresses,
        initial_agent_state=actor_model.initial_state(),
    )

    def run():
        try:
            actors.run()
        except Exception as e:
            logger.error("Exception in actorpool thread!")
            traceback.print_exc()
            raise e

    actorpool_thread = threading.Thread(target=run, name="actorpool-thread", daemon=True)

    optimizer = torch.optim.RMSprop(
        model.parameters(),
        lr=flags.learning_rate,
        momentum=flags.momentum,
        eps=flags.epsilon,
        alpha=flags.alpha,
    )

    def lr_lambda(epoch):
        return 1 - min(epoch * flags.unroll_length * flags.batch_size, flags.total_steps) / flags.total_steps

    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

    stats = {}

    # Load state from a checkpoint, if it exists.
    run_path = "/".join([wandb.run.entity, wandb.run.project, flags.xpid])
    load_checkpoint_filename = flags.load_checkpoint_filename or get_latest_checkpoint_filename(
        run_path, prefix="model_step_", suffix=".tar"
    )
    if load_checkpoint_filename is not None:
        checkpoint = wandb.restore(load_checkpoint_filename, run_path=run_path, replace=True)
        checkpoint_states = torch.load(checkpoint.name, map_location=flags.learner_device)

        model.load_state_dict(checkpoint_states["model_state_dict"])
        optimizer.load_state_dict(checkpoint_states["optimizer_state_dict"])
        scheduler.load_state_dict(checkpoint_states["scheduler_state_dict"])
        stats = checkpoint_states["stats"]
        logger.info(f"Resuming {flags.xpid} from checkpoint {load_checkpoint_filename}, current stats:\n{stats}")
    else:
        assert flags.mode == "train", f"No checkpoint for run {run_path}, cannot test!"

    # Initialize actor model like learner model.
    actor_model.load_state_dict(model.state_dict())

    learner_threads = [
        threading.Thread(
            target=learn,
            name="learner-thread-%i" % i,
            args=(
                flags,
                learner_queue,
                model,
                actor_model,
                optimizer,
                scheduler,
                stats,
            ),
        )
        for i in range(flags.num_learner_threads)
    ]
    inference_threads = [
        threading.Thread(
            target=inference,
            name="inference-thread-%i" % i,
            args=(flags, inference_batcher, actor_model),
        )
        for i in range(flags.num_inference_threads)
    ]
    if flags.mode == "test":
        num_test_worlds = get_num_test_worlds_from_flags(flags)
    else:
        num_test_worlds = 0

    actorpool_thread.start()
    for t in learner_threads + inference_threads:
        t.start()

    def checkpoint(step) -> None:
        if flags.disable_checkpoint:
            return
        checkpointpath = Path(wandb.run.dir) / f"model_step_{step}.tar"
        logger.info(f"Saving checkpoint to {checkpointpath}")
        torch.save(
            {
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
                "scheduler_state_dict": scheduler.state_dict(),
                "stats": stats,
                "flags": vars(flags),
            },
            checkpointpath,
        )
        wandb.save(str(checkpointpath), policy="now")
        wandb.save(str(CURRICULUM_BASE_PATH / "*"), policy="now", base_path=str(CURRICULUM_BASE_PATH.parent))

    test_data = {}
    try:
        last_validation_time = timeit.default_timer()
        last_sample_time = timeit.default_timer()
        while True:
            start_time = timeit.default_timer()
            start_step = stats.get("step", 0)
            if flags.mode == "train" and start_step >= flags.total_steps:
                break
            time.sleep(flags.logging_rate)
            end_step = stats.get("step", 0)

            if flags.mode == "train":
                log_data = stats.copy()
                if (
                    flags.checkpoint_rate > 0
                    and timeit.default_timer() - last_validation_time > 60 * flags.checkpoint_rate
                ):
                    # match up checkpoints with validation, run every checkpoint_rate
                    checkpoint(end_step)
                    last_validation_time = timeit.default_timer()
                logger.info(
                    f"Step {end_step} @ {(end_step - start_step) / (timeit.default_timer() - start_time):.1f} SPS. "
                    f"Inference batcher size: {inference_batcher.size()}. Learner queue size: {learner_queue.size()}."
                    f" Other stats: {stats}"
                )
                wandb.log(log_data)

                if end_step > start_step:
                    last_sample_time = timeit.default_timer()
                elif timeit.default_timer() - last_sample_time > 600:
                    logger.warning("No samples for 10 minutes! Exiting training!")
                    break
            else:
                num_goal_progress_files = len(get_goal_progress_files())
                logger.info(f"{num_goal_progress_files} of {num_test_worlds} test worlds completed")
                # aggressively trying to get these uploaded...
                error_logs = list(Path(GODOT_ERROR_LOG_PATH).glob("*.tar"))
                if len(error_logs) > 0:
                    logger.error(f"Godot errors detected! Uploading and quitting: {error_logs}")
                    for error_log in error_logs:
                        wandb.save(str(error_log), base_path=GODOT_ERROR_LOG_PATH, policy="now")
                    break
                if num_goal_progress_files >= num_test_worlds:
                    break
    except KeyboardInterrupt:
        pass  # Close properly.
    else:
        step = stats["step"]
        if flags.mode == "train":
            logger.info(f"Learning finished after {step} steps. Checkpointing.")
            test_data = stats
            checkpoint(step)
        else:
            logger.info(f"Testing finished! Collecting scores...")
            test_data = get_avalon_test_scores()
            test_data["step"] = step
            wandb.log(test_data)

            test_data["config/total_steps"] = flags.total_steps
            logger.info(BIG_SEPARATOR)
            logger.info(RESULT_TAG)
            logger.info(test_data)
            logger.info(BIG_SEPARATOR)
            if flags.save_test_videos or flags.save_test_actions:
                for log_dir in TORCHBEAST_ENV_LOGS_PATH.glob("*"):
                    if not log_dir.is_dir():
                        continue
                    tar_path = log_dir.with_suffix(".tar.gz")
                    with tarfile.open(tar_path, "w:gz") as tar:
                        tar.add(log_dir, arcname=TORCHBEAST_ENV_LOGS_PATH)
                    wandb.save(str(tar_path), base_path=str(TORCHBEAST_ENV_LOGS_PATH), policy="now")

    # Done with learning. Stop all the ongoing work.
    logger.info("Closing queues and joining threads")
    inference_batcher.close()
    learner_queue.close()
    actorpool_thread.join()
    for t in learner_threads + inference_threads:
        t.join()
    logger.info("Learning cleaned up.")

    return test_data


def main(flags):
    flags.use_lstm = not flags.no_lstm
    if not flags.pipes_basename.startswith("unix:"):
        raise Exception("--pipes_basename has to be of the form unix:/some/path.")

    setup_sentry(percent_of_traces_to_capture=1.0)

    os.makedirs(GODOT_ERROR_LOG_PATH, exist_ok=True)
    wandb.save(
        f"{GODOT_ERROR_LOG_PATH}/*", base_path=GODOT_ERROR_LOG_PATH
    )  # Should watch for anything here and upload immediately?

    if not flags.disable_cuda and torch.cuda.is_available():
        logger.info("Using CUDA.")
        flags.learner_device = torch.device("cuda:0")
        actor_device_index = "1" if torch.cuda.device_count() > 1 else "0"
        flags.actor_device = torch.device(f"cuda:{actor_device_index}")
    else:
        logger.info("Not using CUDA.")
        flags.learner_device = torch.device("cpu")
        flags.actor_device = torch.device("cpu")

    if flags.max_learner_queue_size is None:
        flags.max_learner_queue_size = flags.batch_size

    if flags.write_profiler_trace:
        logger.info("Running with profiler.")
        with torch.autograd.profiler.profile() as prof:
            result = train(flags)
        filename = "chrome-%s.trace" % time.strftime("%Y%m%d-%H%M%S")
        logger.info("Writing profiler trace to '%s.gz'", filename)
        prof.export_chrome_trace(filename)
        os.system("gzip %s" % filename)
    else:
        result = train(flags)

    return result


if __name__ == "__main__":
    flags = parser.parse_args()
    main(flags)



import attr
from loguru import logger

from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.test_envs import TestEnvironmentParams
from avalon.agent.common.trainer import OnPolicyTrainer
from avalon.agent.common.util import setup_new_process
from avalon.agent.ppo.params import PPOParams
from avalon.common.log_utils import configure_remote_logger


@attr.s(auto_attribs=True, frozen=True)
class TestPPOParams(PPOParams):
    total_env_steps: int = 100_000
    num_steps: int = 50
    num_workers: int = 4
    batch_size: int = 50 * 4  # must equal num_steps * num_workers
    discount: float = 0.75
    env_params: TestEnvironmentParams = TestEnvironmentParams(long_episode_length=500000)
    log_freq_hist: int = 10
    log_freq_scalar: int = 1
    log_freq_media: int = 25
    entropy_coef: float = 1e-3  # with a small entropy penalty, random actions should stay random unless useful

    __test__ = False


def run(params: TestPPOParams) -> None:
    assert params.num_batches == 1
    assert params.batch_size == params.num_steps * params.num_workers

    trainer = OnPolicyTrainer(params)
    try:
        trainer.train()
    finally:
        trainer.shutdown()


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = TestPPOParams()
        default_params = parse_args(default_params)
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np
from godot_parser import GDObject
from godot_parser import Node as GDNode
from godot_parser.tree import Tree
from numpy.random import Generator
from scipy.spatial import Delaunay

from avalon.common.utils import first
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.world_creation.configs.flora import FloraConfig
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.debug_plots import plot_points
from avalon.datagen.world_creation.debug_plots import plot_triangulation
from avalon.datagen.world_creation.entities.scenery import Scenery
from avalon.datagen.world_creation.region import EdgedRegion
from avalon.datagen.world_creation.region import Region
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point2DListNP
from avalon.datagen.world_creation.types import RawGDObject
from avalon.datagen.world_creation.utils import ARRAY_MESH_TEMPLATE
from avalon.datagen.world_creation.worlds.biome_map import BiomeMap
from avalon.datagen.world_creation.worlds.height_map import get_flora_config_by_file


class Terrain:
    def __init__(
        self,
        biome_map: BiomeMap,
        is_climbable: MapBoolNP,
        is_detail_important: MapBoolNP,
        point_density_in_points_per_square_meter: float,
        rand: np.random.Generator,
        special_regions: Tuple[Region, ...] = tuple(),
    ) -> None:
        self.biome_map = biome_map
        self.height_map = biome_map.map
        self.fine_triangulation = TerrainTriangulation(
            biome_map,
            is_climbable,
            point_density_in_points_per_square_meter,
            rand,
            special_regions,
            extra_points=biome_map.create_extra_height_points(rand, is_climbable, is_detail_important),
        )
        self.coarse_triangulation = TerrainTriangulation(
            biome_map, is_climbable, point_density_in_points_per_square_meter / 4.0, rand, special_regions
        )

    # TODO: reorganize the order of calls to make a bit more sense
    def export(
        self,
        fine_output_path: Path,
        coarse_output_path: Path,
        region: Region,
        tile_x: int,
        tile_z: int,
        tile_radius: int,
        building_names: List[str],
        trees_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]],
        flora_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]],
        distant_tile_ids: Set[Tuple[int, int]],
        neighboring_and_this_tile_ids: Set[Tuple[int, int]],
    ):
        (
            global_triangle_indices_for_region,
            region_triangles,
            region_triangle_normals,
            region_vertex_normals,
            region_vertices,
            region_colors,
        ) = self._get_export_data_for_region(region)

        scene = GodotScene()

        with scene.use_tree() as tree:
            tree.root = _create_terrain_info_node(
                scene,
                tile_x,
                tile_z,
                building_names,
            )

            tree.root.add_child(
                _create_terrain_mesh_node(
                    scene,
                    region_triangles,
                    region_vertices,
                    region_vertex_normals,
                    region_colors,
                    self.biome_map.config.is_terrain_noise_shader_enabled,
                    is_fine=True,
                )
            )

            # add trees for this tile
            our_tile_id_set = {(tile_z, tile_x)}
            tree_list_by_resource = get_scenery_with_tile_ids(trees_by_tile_ids_and_resource, our_tile_id_set)
            for tree_resource, tree_list in tree_list_by_resource.items():
                first_tree = first(tree_list)
                assert first_tree is not None
                tree.root.add_child(
                    create_multimesh_instance(
                        scene,
                        f"high_tree_{tile_x}_{tile_z}",
                        tree_list,
                        first_tree.resource_file.replace(".tscn", ".res"),
                        self.biome_map.config.flora_config,
                    )
                )

            # if this is a single-tile export scenario, include the flora in here
            if len(neighboring_and_this_tile_ids) == 1:
                self._add_tile_local_flora(
                    flora_by_tile_ids_and_resource, neighboring_and_this_tile_ids, scene, tile_x, tile_z, tree
                )

        scene.write(str(fine_output_path.absolute()))

        # create the file that shows the whole rest of the world
        distant_export_data = self._get_export_data_outside_of_region(region, tile_radius)

        scene = GodotScene()
        with scene.use_tree() as tree:
            if distant_export_data is not None:
                (
                    selected_triangle_indices,
                    distant_triangles,
                    distant_triangle_normals,
                    distant_vertex_normals,
                    distant_vertices,
                    distant_colors,
                ) = distant_export_data

                tree.root = _create_terrain_mesh_node(
                    scene,
                    distant_triangles,
                    distant_vertices,
                    distant_vertex_normals,
                    distant_colors,
                    self.biome_map.config.is_terrain_noise_shader_enabled,
                    is_fine=False,
                )
            else:
                tree.root = GDNode(
                    "DistantData",
                    type="Spatial",
                    properties={"transform": GDObject("Transform", 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0)},
                )

            tree_list_by_resource = get_scenery_with_tile_ids(trees_by_tile_ids_and_resource, distant_tile_ids)
            for tree_resource, tree_list in tree_list_by_resource.items():
                first_tree = first(tree_list)
                assert first_tree is not None
                tree.root.add_child(
                    create_multimesh_instance(
                        scene,
                        f"high_tree_{tile_x}_{tile_z}",
                        tree_list,
                        first_tree.resource_file.replace(".tscn", "_low.res"),
                        self.biome_map.config.flora_config,
                    )
                )

            if len(neighboring_and_this_tile_ids) > 1:
                # get the other scenery from the neighboring tiles and combine with our own
                self._add_tile_local_flora(
                    flora_by_tile_ids_and_resource, neighboring_and_this_tile_ids, scene, tile_x, tile_z, tree
                )

        scene.write(str(coarse_output_path.absolute()))

        return region_triangles, region_vertices

    def _add_tile_local_flora(
        self,
        flora_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]],
        neighboring_and_this_tile_ids: Set[Tuple[int, int]],
        scene: GodotScene,
        tile_x: int,
        tile_z: int,
        tree: Tree,
    ) -> None:
        flora_list_by_resource = get_scenery_with_tile_ids(
            flora_by_tile_ids_and_resource, neighboring_and_this_tile_ids
        )
        for flora_resource, flora_list in flora_list_by_resource.items():
            first_flora = first(flora_list)
            assert first_flora is not None
            tree.root.add_child(
                create_multimesh_instance(
                    scene,
                    f"high_tree_{tile_x}_{tile_z}",
                    flora_list,
                    first_flora.resource_file.replace(".tscn", ".res"),
                    self.biome_map.config.flora_config,
                )
            )

    def _get_export_data_for_region(
        self, region: Region
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        global_triangle_indices = self._are_triangles_in_tile(
            self.fine_triangulation.triangle_points, region
        ).nonzero()[0]

        result = _get_export_data_inner(global_triangle_indices, self.fine_triangulation)
        assert result is not None, "No triangles in region?"
        return result

    def _get_export_data_outside_of_region(
        self, region: Region, num_tiles: int
    ) -> Union[None, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
        x_min = region.x.min_ge - num_tiles * region.x.size
        x_max = x_min + region.x.size * (2 * num_tiles + 1)
        z_min = region.z.min_ge - num_tiles * region.z.size
        z_max = z_min + region.z.size * (2 * num_tiles + 1)
        loaded_region = Region(
            x=FloatRange(x_min, x_max),
            z=FloatRange(z_min, z_max),
        )

        global_triangle_indices = self._are_triangles_outside_tile(
            self.coarse_triangulation.triangle_points, loaded_region
        ).nonzero()[0]

        return _get_export_data_inner(global_triangle_indices, self.coarse_triangulation)

    def _are_triangles_in_tile(self, triangle_points: np.ndarray, region: Region) -> np.ndarray:
        """Returns a boolean numpy array indicating the triangles that fall in the region"""
        return cast(
            np.ndarray,
            (
                (triangle_points[:, :, 0] >= region.x.min_ge).any(axis=-1)
                & (triangle_points[:, :, 0] < region.x.max_lt).any(axis=-1)
                & (triangle_points[:, :, 2] >= region.z.min_ge).any(axis=-1)
                & (triangle_points[:, :, 2] < region.z.max_lt).any(axis=-1)
            ),
        )

    def _are_triangles_outside_tile(self, triangle_points: np.ndarray, region: Region) -> np.ndarray:
        """Returns a boolean numpy array indicating the triangles that fall outside of region"""
        return cast(
            np.ndarray,
            (
                (triangle_points[:, :, 0] < region.x.min_ge).any(axis=-1)
                | (triangle_points[:, :, 0] >= region.x.max_lt).any(axis=-1)
                | (triangle_points[:, :, 2] < region.z.min_ge).any(axis=-1)
                | (triangle_points[:, :, 2] >= region.z.max_lt).any(axis=-1)
            ),
        )


#
# def create_triset(
#     geom: Geometry, input_list: InputList, triangle_indices: np.ndarray, material_ref: str
# ) -> TriangleSet:
#     flat_triangle_indices = triangle_indices.flatten()
#     flat_triangle_indices = np.reshape(flat_triangle_indices, (len(flat_triangle_indices), 1))
#     flat_normal_indices = flat_triangle_indices
#     indices = np.concatenate((flat_triangle_indices, flat_normal_indices), axis=1).flatten()
#     triset = geom.createTriangleSet(indices, input_list, material_ref)
#     return triset


def normalize_v3(arr: np.ndarray) -> np.ndarray:
    """Normalize a numpy array of 3 component vectors shape=(n,3)"""
    lens = np.sqrt(arr[:, 0] ** 2 + arr[:, 1] ** 2 + arr[:, 2] ** 2)
    arr[:, 0] /= lens
    arr[:, 1] /= lens
    arr[:, 2] /= lens
    return arr


# roughly from here: https://sites.google.com/site/dlampetest/python/calculating-normals-of-a-triangle-mesh-using-numpy
def calculate_normals(vertices: np.ndarray, faces: np.ndarray) -> np.ndarray:
    # Create an indexed view into the vertex array using the array of three indices for triangles
    tris = vertices[faces]

    # Calculate the normal for all the triangles, by taking the cross product of the vectors v1-v0, and v2-v0 in each triangle
    n = np.cross(tris[::, 1] - tris[::, 0], tris[::, 2] - tris[::, 0]) * -1.0

    # n is now an array of normals per triangle. The length of each normal is dependent the vertices,
    # we need to normalize these, so that our next step weights each normal equally.
    normalize_v3(n)

    return cast(np.ndarray, n)


def calculate_normals_for_faces_and_vertices(vertices: np.ndarray, faces: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    # Create a zeroed array with the same type and shape as our vertices i.e., per vertex normal
    norm = np.zeros(vertices.shape, dtype=vertices.dtype)

    # Create an indexed view into the vertex array using the array of three indices for triangles
    tris = vertices[faces]

    # Calculate the normal for all the triangles, by taking the cross product of the vectors v1-v0, and v2-v0 in each triangle
    n = np.cross(tris[::, 1] - tris[::, 0], tris[::, 2] - tris[::, 0])

    # n is now an array of normals per triangle. The length of each normal is dependent the vertices,
    # we need to normalize these, so that our next step weights each normal equally.
    normalize_v3(n)

    # now we have a normalized array of normals, one per triangle, i.e., per triangle normals.
    # But instead of one per triangle (i.e., flat shading), we add to each vertex in that triangle,
    # the triangles' normal. Multiple triangles would then contribute to every vertex, so we need to normalize again afterwards.
    # The cool part, we can actually add the normals through an indexed view of our (zeroed) per vertex normal array
    norm[faces[:, 0]] += n
    norm[faces[:, 1]] += n
    norm[faces[:, 2]] += n
    # normalize_v3(norm)
    # so I replaced with this for now, which is significantly slower (10%)
    lens = np.sqrt(norm[:, 0] ** 2 + norm[:, 1] ** 2 + norm[:, 2] ** 2)
    norm[:, 0] = np.divide(norm[:, 0], lens, out=np.zeros_like(lens), where=lens != 0)
    norm[:, 1] = np.divide(norm[:, 1], lens, out=np.ones_like(lens), where=lens != 0)
    norm[:, 2] = np.divide(norm[:, 2], lens, out=np.zeros_like(lens), where=lens != 0)

    return norm, n


def _create_triangulation(
    rand: Generator,
    edge_region: EdgedRegion,
    point_density_in_points_per_square_meter: float,
    special_regions: Tuple[Region, ...] = tuple(),
    extra_points: Optional[np.ndarray] = None,
) -> Delaunay:
    """
    Create random points within a region and Delaunay triangulate them to create a Terrain object

    Helper for Terrain
    """
    area = edge_region.x.size * edge_region.z.size
    num_points = round(area * point_density_in_points_per_square_meter)
    points = edge_region.get_randomish_points(num_points)

    if IS_DEBUG_VIS:
        plot_points(points, 0, 1)

    if extra_points is not None:
        points = np.concatenate([points, extra_points])

    if IS_DEBUG_VIS:
        plot_points(points, 0, 1)

    for region in special_regions:
        assert edge_region.contains_region(
            region
        ), f"Some special regions are not contained in the tile region: {region}"
        points = np.array([x for x in points if not region.contains_point_2d(x)] + region.points)
    edge_points = np.array([(x[0], x[2]) for x in edge_region.edge_vertices])
    points = np.concatenate((edge_points, points))
    if IS_DEBUG_VIS:
        plot_points(points, 0, 1)
    return Delaunay(points)


class TerrainTriangulation:
    def __init__(
        self,
        biome_map: BiomeMap,
        is_climbable: MapBoolNP,
        point_density_in_points_per_square_meter: float,
        rand: np.random.Generator,
        special_regions: Tuple[Region, ...] = tuple(),
        extra_points: Optional[Point2DListNP] = None,
    ) -> None:
        self.config = biome_map.config
        height_map = biome_map.map
        points_per_meter = point_density_in_points_per_square_meter**0.5
        edge_region = height_map.create_edge_region(height_map.region, points_per_meter, rand)
        triangulation = _create_triangulation(
            rand,
            edge_region,
            point_density_in_points_per_square_meter,
            special_regions,
            extra_points,
        )
        self.neighbors = triangulation.neighbors
        if IS_DEBUG_VIS:
            plot_triangulation(triangulation)
        heights = height_map.get_heights(triangulation.points)
        height_by_id = {i: heights[i] for i in range(len(triangulation.points))}
        # reset the edge vertices to whatever heights are specified
        for i, vertex in enumerate(edge_region.edge_vertices):
            height_by_id[i] = vertex[1]

        # convert to 3D points from Delaunay triangulation
        self.vertices = np.array([(x[0], height_by_id[i], x[1]) for i, x in enumerate(triangulation.points)])

        self.triangles = triangulation.simplices

        self.triangle_points = self.vertices[self.triangles]

        # only calculate vertex normals if required
        if biome_map.config.is_normal_per_vertex or biome_map.config.is_color_per_vertex:
            vertex_normals, triangle_normals = calculate_normals_for_faces_and_vertices(self.vertices, self.triangles)
            vertex_normals *= -1

        # set per vertex or per face normals, as configured
        if biome_map.config.is_normal_per_vertex:
            self.vertex_normals, self.triangle_normals = vertex_normals, triangle_normals
        else:
            self.triangle_normals = calculate_normals(self.vertices, self.triangles)

        self.midpoints = np.array([self.vertices[x].sum(axis=0) / 3.0 for x in self.triangles])

        # set per vertex or per face colors, as configured
        if biome_map.config.is_color_per_vertex:
            self.vertex_colors = biome_map.get_colors(rand, is_climbable, self.vertices, vertex_normals)
        else:
            self.triangle_colors = biome_map.get_colors(rand, is_climbable, self.midpoints, self.triangle_normals)


def get_scenery_with_tile_ids(
    scenery_by_tile_ids_and_resource: Dict[Tuple[int, int], Dict[str, List[Scenery]]], tile_ids: Set[Tuple[int, int]]
) -> Dict[str, List[Scenery]]:
    results: Dict[str, List[Scenery]] = {}
    for tile_id, inner_dict in scenery_by_tile_ids_and_resource.items():
        if tile_id in tile_ids:
            for resource, scenery in inner_dict.items():
                if resource not in results:
                    results[resource] = []
                results[resource].extend(scenery)
    return results


def _get_export_data_inner(
    global_triangle_indices: np.ndarray, triangulation: TerrainTriangulation
) -> Union[None, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
    if len(global_triangle_indices) == 0:
        return None
    triangles = triangulation.triangles[global_triangle_indices]
    triangle_normals = triangulation.triangle_normals[global_triangle_indices]

    is_color_per_vertex = triangulation.config.is_color_per_vertex
    is_normal_per_vertex = triangulation.config.is_normal_per_vertex

    new_triangles = np.array([(3 * i, 3 * i + 1, 3 * i + 2) for i in range(len(triangles))])
    vertices = []
    colors = []
    normals = []
    for triangle_index in global_triangle_indices:
        triangle = triangulation.triangles[triangle_index]
        for vertex in triangle:
            vertices.append(triangulation.vertices[vertex])
            # colors.append(triangulation.triangle_colors[triangle_index])
            # normals.append(triangulation.triangle_normals[triangle_index])
            colors.append(
                triangulation.vertex_colors[vertex]
                if is_color_per_vertex
                else triangulation.triangle_colors[triangle_index]
            )
            normals.append(
                triangulation.vertex_normals[vertex]
                if is_normal_per_vertex
                else triangulation.triangle_normals[triangle_index]
            )

    return (
        global_triangle_indices,
        new_triangles,
        triangle_normals,
        np.array(normals),
        np.array(vertices),
        np.array(colors),
    )


def _create_terrain_mesh_node(
    scene: GodotScene,
    material_triangles: np.ndarray,
    vertices: np.ndarray,
    vertex_normals: np.ndarray,
    colors: np.ndarray,
    is_terrain_noise_shader_enabled: bool,
    is_fine: bool,
) -> GDNode:
    surface_kwargs = {}
    surface_id = 0
    if is_terrain_noise_shader_enabled:
        material_resource = scene.add_ext_resource("res://materials/noise_terrain.material", "Material")
    else:
        material_resource = scene.add_ext_resource("res://materials/terrain.material", "Material")
    aabb_corner = [x.min() for x in [vertices[:, 0], vertices[:, 1], vertices[:, 2]]]
    aabb_maxes = [x.max() for x in [vertices[:, 0], vertices[:, 1], vertices[:, 2]]]
    aabb_size = tuple(aabb_maxes[i] - aabb_corner[i] for i in range(3))
    aabb = (*aabb_corner, *aabb_size)
    alpha_colors = np.concatenate([colors, np.ones((len(colors), 1))], axis=1)
    surface_kwargs[f"surfaces/{surface_id}"] = RawGDObject(
        ARRAY_MESH_TEMPLATE.format(
            aabb=", ".join(str(x) for x in aabb),
            vertex_floats=_print_array(vertices.flatten(), 3),
            vertex_normal_floats=_print_array(vertex_normals.flatten(), 3),
            color_floats=_print_array(alpha_colors.flatten(), 5),
            triangle_indices=_print_array(material_triangles.flatten(), 1),
            index_count=str(len(material_triangles)),
            vertex_count=str(len(vertices)),
            mesh_name="terrain",
            material_resource_type="ExtResource",
            material_id=material_resource.id,
        )
    )
    surface_id += 1
    terrain_mesh = scene.add_sub_resource("ArrayMesh", resource_name="terrain", **surface_kwargs)
    terrain_mesh_node = GDNode(
        "terrain_mesh" if is_fine else "distant_mesh",
        type="MeshInstance",
        properties={"mesh": terrain_mesh.reference, "material/0": "null"},
    )
    return terrain_mesh_node


def _create_static_body(index: int, scene: GodotScene, triangles: np.ndarray, vertices: np.ndarray) -> GDNode:

    collision_mesh_vertices = []
    for triangle in triangles:
        collision_mesh_vertices.append(vertices[triangle[0]])
        collision_mesh_vertices.append(vertices[triangle[1]])
        collision_mesh_vertices.append(vertices[triangle[2]])
    collision_mesh_floats = _print_array(np.array(collision_mesh_vertices).flatten(), 3)
    collision_shape = scene.add_sub_resource(
        "ConcavePolygonShape",
        data=RawGDObject(
            "PoolVector3Array( {collision_mesh_floats} )".format(collision_mesh_floats=collision_mesh_floats)
        ),
    )
    collision_shape_node = GDNode(
        f"static_collision_shape_{index}", type="CollisionShape", properties={"shape": collision_shape.reference}
    )
    physics_material = scene.add_sub_resource("PhysicsMaterial", bounce=0.0, absorbent="false")
    static_body = GDNode(
        f"static_collision_{index}",
        type="StaticBody",
        properties={"physics_material_override": physics_material.reference},
    )
    static_body.add_child(collision_shape_node)
    return static_body


def _create_terrain_info_node(
    scene: GodotScene,
    tile_x: int,
    tile_z: int,
    building_names: List[str],
) -> GDNode:
    quoted_building_names = [f'"{x}"' for x in building_names]
    info_node = GDNode(
        "terrain",
        type="Spatial",
        properties={
            "script": scene.add_ext_resource("res://terrain/terrain_tile.gd", "Script").reference,
            "tile_x": tile_x,
            "tile_z": tile_z,
            "building_names": RawGDObject("[ " + ", ".join(quoted_building_names) + " ]"),
        },
    )
    return info_node


def _print_array(array: np.ndarray, precision: int) -> str:
    # annoyingly, this is MUCH slower than the below...
    # return np.array2string(
    #     array,
    #     threshold=np.inf,
    #     precision=precision,
    #     max_line_width=np.inf,
    #     suppress_small=True,
    #     floatmode="fixed",
    #     separator=", ",
    # )[1:-1]
    return ", ".join([f"{x:.{precision}f}" for x in array.flatten()])


def create_multimesh_instance(
    scene: GodotScene,
    name: str,
    instances: list[Scenery],
    resource_file: str,
    flora_config: Dict[str, FloraConfig],
):

    config, resource_name = get_flora_config_by_file(flora_config, resource_file)
    assert config is not None
    if config.is_noise_shader_enabled:
        material_resource = scene.add_ext_resource("res://materials/flora.material", "Material")
    else:
        material_resource = None

    transform_data = np.array(
        [
            np.array([*((x.rotation.reshape((3, 3)) * x.scale * config.default_scale).flatten()), *x.position])
            for x in instances
        ]
    ).flatten()

    transform_floats = _print_array(np.array(transform_data).flatten(), 3)

    mesh_resource = scene.add_ext_resource(resource_file, "ArrayMesh")
    multimesh_resource = scene.add_sub_resource(
        "MultiMesh",
        transform_format=1,
        instance_count=len(instances),
        mesh=mesh_resource.reference,
        transform_array=RawGDObject(
            "PoolVector3Array( {transform_floats} )".format(transform_floats=transform_floats)
        ),
    )
    return GDNode(
        f"{name}_{resource_name}",
        type="MultiMeshInstance",
        properties={
            "transform": GDObject("Transform", 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0),
            "multimesh": multimesh_resource.reference,
            "material_override": material_resource.reference if material_resource is not None else None,
            "cast_shadow": 1 if config.is_shadowed else 0,
        },
    )



import attr
import numpy as np
from godot_parser import GDObject
from godot_parser import Node as GDNode
from godot_parser import Vector3
from scipy.spatial.transform import Rotation

from avalon.datagen.world_creation.entities.item import Item
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Pillar(Item):
    size: np.ndarray
    is_dynamic: bool = False
    yaw: float = 0.0
    entity_id: int = -1

    def get_node(self, scene: GodotScene) -> GDNode:
        yaw_rotation = Rotation.from_euler("y", self.yaw, degrees=True)
        rotation = yaw_rotation.as_matrix().flatten()
        transform = GDObject("Transform", *rotation, *self.position)
        root = GDNode(
            f"pillar_{self.entity_id}",
            type="StaticBody",
            properties={
                "transform": transform,
            },
        )
        root.add_child(
            GDNode(
                "mesh",
                type="MeshInstance",
                properties={
                    "mesh": scene.add_sub_resource(
                        "CubeMesh",
                        material=scene.add_ext_resource("res://shaders/BasicColor.material", "Material").reference,
                        size=Vector3(*self.size),
                    ).reference,
                    "material/0": None,
                },
            )
        )
        root.add_child(
            GDNode(
                "collision",
                type="CollisionShape",
                properties={
                    "shape": scene.add_sub_resource(
                        "BoxShape", resource_local_to_scene=True, extents=Vector3(*self.size / 2.0)
                    ).reference
                },
            )
        )

        return root



# mypy: ignore-errors
# We ignore all types in this file because typing in `gym==0.25.*` is a mess.
# Will fix when we update gym to `>=0.26` or wait for the `1.0` release.

import copy
from collections import defaultdict
from collections import deque
from typing import Any
from typing import Callable
from typing import Deque
from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import gym
import numpy as np
import torch
from gym import Wrapper
from gym import spaces
from gym.spaces import Box
from loguru import logger
from numpy.typing import DTypeLike
from numpy.typing import NDArray
from torch.nn import functional as F
from torchvision import transforms
from tree import map_structure

# All wrappers should operate only on ndarrays, no torch Tensors!
ArrayType = np.typing.NDArray
DictActionType = dict[str, ArrayType]
DictObservationType = dict[str, ArrayType]

GenericObservationType = TypeVar("GenericObservationType")
GenericActionType = TypeVar("GenericActionType")


class DictObsActionWrapper(Wrapper[DictObservationType, DictActionType]):
    """Give dictionary observations, and take dictionary actions."""

    def __init__(
        self,
        env: gym.Env[Union[DictObservationType, NDArray], Union[DictObservationType, NDArray]],
        obs_key: str = "wrapped",
        action_key: str = "wrapped",
    ) -> None:
        super().__init__(env)

        # Build observation space
        self.wrapped_obs = False
        if not isinstance(self.env.observation_space, gym.spaces.Dict):
            self.observation_space = gym.spaces.Dict({obs_key: self.env.observation_space})
            self.wrapped_obs = True

        self.wrapped_action = False
        self.action_space: gym.spaces.Space
        if not isinstance(self.action_space, gym.spaces.Dict):
            self.action_space = gym.spaces.Dict({action_key: self.action_space})
            self.wrapped_action = True

        self._env = env
        self.obs_key = obs_key
        self.action_key = action_key

    def reset(self, *args, **kwargs) -> DictObservationType:  # type: ignore[no-untyped-def]
        observation = self.env.reset(*args, **kwargs)
        return self.observation(observation)

    def step(self, action: DictActionType) -> tuple[DictObservationType, float, bool, dict]:
        unwrapped_action = self.action(action)
        observation, reward, done, info = self.env.step(unwrapped_action)
        return self.observation(observation), reward, done, info

    def action(self, action: DictActionType) -> Union[NDArray, DictActionType]:
        if self.wrapped_action:
            return action[self.action_key]
        else:
            return action

    def observation(self, observation: Union[NDArray, DictObservationType]) -> DictObservationType:
        if self.wrapped_obs:
            assert isinstance(observation, np.ndarray)
            return {self.obs_key: observation}
        else:
            return observation


class PixelObsWrapper(gym.ObservationWrapper):
    """Render state-based envs to pixels and use that as the observation."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        pixels: NDArray = self.env.render(mode="rgb_array")
        # Some envs (eg CartPole-v1) won't render without being reset first.
        if pixels is None:
            self.env.reset()
            pixels = self.env.render(mode="rgb_array")
        self.observation_space = spaces.Box(shape=pixels.shape, low=0, high=255, dtype=pixels.dtype)
        self._env = env

    def observation(self, observation: NDArray) -> NDArray:
        obs: NDArray = self.env.render(mode="rgb_array")
        return obs


class ImageTransformWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env, key: str, resolution: Optional[int] = None, greyscale: bool = False) -> None:
        """Transform a uint8 image into the format expected by a model.

        We actually leave images as uint8s until training, for the 4x lower data size."""
        super().__init__(env)
        self._env = env
        noop = transforms.Lambda(lambda x: x)
        self.transform = transforms.Compose(
            [
                # input is numpy (h w c)
                transforms.ToPILImage(),
                transforms.Grayscale() if greyscale else noop,
                transforms.Resize((resolution, resolution)) if resolution else noop,
                # transforms.ToTensor(), # this converts hwc to chw, and from 0-255 to 0-1
                transforms.PILToTensor(),  # this converts hwc to chw, without scaling
                # Convert that range to [-0.5, 0.5]
                # transforms.Lambda(lambda x: x - 0.5),
            ]
        )

        observation_space = env.observation_space
        assert isinstance(observation_space, spaces.Dict)
        space = observation_space[key]

        assert isinstance(space, spaces.Box)
        assert np.all(space.high == 255)
        assert np.all(space.low == 0)
        assert space.dtype == np.uint8, space.dtype
        assert len(space.shape) == 3
        if space.shape[2] in (1, 3):
            self.channels = space.shape[2] if not greyscale else 1
        else:
            assert False
        self.res = resolution if resolution else space.shape[1]
        # new_space = spaces.Box(shape=(self.channels, self.res, self.res), low=-0.5, high=0.5, dtype=np.float32)
        new_space = spaces.Box(shape=(self.channels, self.res, self.res), low=0, high=255, dtype=np.uint8)
        self.observation_space = copy.copy(observation_space)
        self.observation_space.spaces[key] = new_space
        self.key = key

    def observation(self, obs: Dict[str, NDArray]) -> Dict[str, NDArray]:
        image = obs[self.key]
        image = self.transform(image).numpy()
        assert image.shape == (self.channels, self.res, self.res), image.shape
        obs[self.key] = image
        return obs


class ClipActionWrapper(gym.ActionWrapper):
    """Clip actions in Box envs that are out of bounds."""

    def __init__(self, env: gym.Env) -> None:
        assert isinstance(env.action_space, gym.spaces.Dict)
        super().__init__(env)

    def action(self, action: DictActionType) -> DictActionType:
        for k, space in self.action_space.spaces.items():
            if isinstance(space, gym.spaces.Box):
                action[k] = np.clip(action[k], space.low, space.high)
        return action

    def reverse_action(self, action: DictActionType) -> DictActionType:
        raise NotImplementedError


class DiscreteActionToIntWrapper(gym.ActionWrapper):
    """Mujoco (maybe all gym envs?) only take ints for discrete actions.
    My code generates (arrays of) floats (this is what torch.distributions.Bernoulli.sample() gives).

    So for now I'll leave them in floats in my code, and convert them in this wrapper."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)

    def action(self, action: np.floating) -> Union[np.floating, int]:
        if isinstance(self.action_space, gym.spaces.Discrete):
            return int(action.item())
        else:
            return action


class OneHotDiscrete(gym.spaces.Discrete):
    def sample(self, mask: Optional[NDArray] = None) -> NDArray:
        assert mask is None
        sample = self.np_random.randint(self.n)
        return F.one_hot(torch.from_numpy(sample), num_classes=self.n).numpy()


class OneHotMultiDiscrete(gym.spaces.MultiDiscrete):
    def __init__(self, nvec: List[int], dtype: DTypeLike = np.float32) -> None:
        # the dtype is default float32 since that's what torch.OneHotCategorical.sample() returns,
        # for some reason (presumably to make backprop work).
        # This is kinda weird tho because then it makes nvec a float32 too.. this superclass really expects an int
        # But we need self.dtype to be the type that we pass around during rollouts,
        # since this is where it gets the dtype to use for storing the actions.
        # One option is to make self.dtype be an integer type, use integers all thru rollouts/inference,
        # and floats in training. Confusing though.
        # Another is to always use float, and just patch this class to make it handle that properly.
        # That's the approach I've gone with for now.
        self.max_categories = max(nvec)
        super().__init__(nvec=nvec, dtype=dtype)

    def sample(self, mask: Optional[NDArray] = None) -> NDArray:
        assert mask is None
        sample = (self.np_random.random(self.nvec.shape) * self.nvec).astype(np.int32)
        return (
            F.one_hot(torch.from_numpy(sample).long(), num_classes=self.max_categories).to(dtype=torch.float32).numpy()
        )


class OneHotActionWrapper(gym.ActionWrapper):
    """Convert all discrete spaces to a OneHotMultiDiscrete space.
    My code will generate one-hot actions with shape (num_actions, num_categories),
    and this wrapper will convert to integer actions of shape necessary for each space type.

    NOTE: This probably won't work quite right for a true MultiDiscrete space with diff number of classes
    for each dimension. Since we'll use max(num_classes) as the one-hot vector length for all.
    """

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        assert isinstance(env.action_space, gym.spaces.Dict)
        self._env = env

        self.action_space = copy.deepcopy(env.action_space)
        for k, space in self.action_space.spaces.items():
            if isinstance(space, gym.spaces.Discrete):
                # self.action_space[k] = OneHotDiscrete(space.n)
                self.action_space.spaces[k] = OneHotMultiDiscrete([space.n])
            if isinstance(space, gym.spaces.MultiDiscrete):
                self.action_space.spaces[k] = OneHotMultiDiscrete(space.nvec)
                assert len(set(space.nvec)) == 1, "not sure we support heterogenous MultiDiscrete spaces properly."
            if isinstance(space, gym.spaces.MultiBinary):
                self.action_space.spaces[k] = OneHotMultiDiscrete([2] * space.n)

    def action(self, action_dict: DictActionType) -> dict[str, Any]:
        out: dict[str, Any] = copy.copy(action_dict)
        for k, space in self._env.action_space.spaces.items():
            action = action_dict[k]
            assert isinstance(action, np.ndarray)

            if isinstance(space, gym.spaces.Discrete):
                assert action.shape == (1, space.n)
                out[k] = np.argmax(action[0]).item()
            elif isinstance(space, gym.spaces.MultiDiscrete):
                # assert action.shape == (len(space.nvec), max(space.nvec))
                out[k] = np.argmax(action, axis=-1)
            elif isinstance(space, gym.spaces.MultiBinary):
                assert action.shape == (space.n, 2)
                out[k] = np.argmax(action, axis=-1)
        return out

    def reverse_action(self, action: Dict[str, NDArray]):
        raise NotImplementedError


class TimeLimit(gym.Wrapper[GenericObservationType, GenericActionType]):
    def __init__(self, env: gym.Env, max_episode_steps: Optional[int] = None) -> None:
        super().__init__(env)
        self._max_episode_steps: Optional[int] = max_episode_steps
        self._elapsed_steps: Optional[int] = None

    def step(self, action: GenericActionType) -> tuple[GenericObservationType, float, bool, dict]:
        assert self._max_episode_steps is not None, "max_episode_steps must be set"
        assert self._elapsed_steps is not None, "Cannot call env.step() before calling reset()"
        observation, reward, done, info = self.env.step(action)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            info["TimeLimit.truncated"] = not done
            done = True
        return observation, reward, done, info

    def reset(self, *args, **kwargs) -> GenericObservationType:
        self._elapsed_steps = 0
        return self.env.reset(*args, **kwargs)


class ElapsedTimeWrapper(gym.Wrapper[DictObservationType, GenericActionType]):
    def __init__(self, env: gym.Env, max_episode_steps: Optional[int] = None) -> None:
        super().__init__(env)
        self._max_episode_steps: Optional[int] = max_episode_steps
        self._elapsed_steps: Optional[int] = None

        assert self._max_episode_steps is not None
        assert isinstance(env.observation_space, gym.spaces.Dict)
        self.observation_space = copy.copy(self.env.observation_space)
        self.observation_space["elapsed_time"] = gym.spaces.Box(low=-1, high=1, shape=(1,), dtype=np.float32)

    def observation(self, observation: DictObservationType) -> DictObservationType:
        assert self._elapsed_steps <= self._max_episode_steps
        observation["elapsed_time"] = np.array(((self._elapsed_steps / self._max_episode_steps) - 0.5) * 2)
        return observation

    def step(self, action: GenericActionType) -> tuple[DictObservationType, float, bool, dict]:
        assert self._elapsed_steps is not None, "Cannot call env.step() before calling reset()"
        observation, reward, done, info = self.env.step(action)
        self._elapsed_steps += 1
        return self.observation(observation), reward, done, info

    def reset(self, *args, **kwargs) -> DictObservationType:
        self._elapsed_steps = 0
        observation = self.env.reset(*args, **kwargs)
        return self.observation(observation)


class NormalizeActions(gym.ActionWrapper):
    """Adapts envs with arbitrary action space bounds to take actions with a standard (-1, 1) range.

    Does not adapt non-Box spaces or dimensions that have non-finite bounds."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self._env = env
        self.bounds = {}
        assert isinstance(env.action_space, gym.spaces.Dict)
        self.action_space = copy.deepcopy(env.action_space)
        for k, space in env.action_space.spaces.items():
            if isinstance(space, gym.spaces.Box):
                mask = np.logical_and(np.isfinite(space.low), np.isfinite(space.high))
                if any(mask == 0):
                    logger.warning(f"Action space '{k}' has non-finite bounds. Action rescaling will not be applied.")
                low = np.where(mask, space.low, -1)
                high = np.where(mask, space.high, 1)
                self.bounds[k] = (mask, low, high)

                new_low = np.where(mask, -np.ones_like(low), low)
                new_high = np.where(mask, np.ones_like(low), high)
                self.action_space[k] = gym.spaces.Box(new_low, new_high, dtype=np.float32)

    def action(self, action: DictActionType) -> DictActionType:
        for k, v in action.items():
            if isinstance(self.action_space[k], gym.spaces.Box):
                mask, low, high = self.bounds[k]
                new = (v + 1) / 2 * (high - low) + low
                action[k] = np.where(mask, new, v)
        return action

    def reverse_action(self, action: DictActionType) -> DictActionType:
        raise NotImplementedError


class ScaleRewards(gym.RewardWrapper):
    def __init__(self, env: gym.Env, scale: float = 1.0, func: Callable[[float], float] = lambda x: x) -> None:
        super().__init__(env)
        self.scale = scale
        self.func = func

    def reward(self, reward: float) -> float:
        return self.func(reward) * self.scale


class ActionRepeat(gym.Wrapper[GenericObservationType, GenericActionType]):
    def __init__(self, env: gym.Env, amount: int) -> None:
        super().__init__(env)
        self._amount = amount

    def step(self, action: GenericActionType) -> tuple[GenericObservationType, float, bool, dict]:
        done = False
        total_reward = 0
        current_step = 0
        while current_step < self._amount and not done:
            obs, reward, done, info = self.env.step(action)
            total_reward += reward
            current_step += 1
        # Note: this will just return the info of the most recent step
        return obs, total_reward, done, info


class DictFrameStack(gym.ObservationWrapper):
    """Frame-stack all observations in a Dict space.

    Shapes will be converted from (c, *) to (c * num_stack, *). So for images, this will only work with (c,h,w).
    For the first steps after a reset, where there aren't `num_stack` prior steps,
    the observation at t=0 will be repeated back in time to fill.
    """

    def __init__(self, env: gym.Env, num_stack: int) -> None:
        super().__init__(env)
        self._env = env
        self.num_stack = num_stack

        assert isinstance(env.observation_space, gym.spaces.Dict)
        new_observation_space: dict[str, gym.Space] = {}
        for k, space in env.observation_space.spaces.items():
            if isinstance(space, gym.spaces.Box):
                low = np.repeat(space.low, num_stack, axis=0)
                high = np.repeat(space.high, num_stack, axis=0)
                new_observation_space[k] = Box(low=low, high=high, dtype=space.dtype)
            else:
                assert False, (k, space)

        self.observation_space = gym.spaces.Dict(new_observation_space)
        self.history: dict[str, Deque] = defaultdict(lambda: deque(maxlen=self.num_stack))

    def add_to_history(self, obs: dict[str, NDArray]) -> None:
        for k, v in obs.items():
            self.history[k].append(v)

    def reset(self, **kwargs):
        self.history: dict[str, Deque] = defaultdict(lambda: deque(maxlen=self.num_stack))
        obs = self.env.reset(**kwargs)
        # We'll initialize the buffer with copies of the first frame
        [self.add_to_history(obs) for _ in range(self.num_stack - 1)]
        return self.observation(obs)

    def observation(self, obs: Dict[str, NDArray]) -> Dict[str, NDArray]:
        self.add_to_history(obs)
        out: dict[str, NDArray] = {}
        for k, v in obs.items():
            assert len(self.history[k]) == self.num_stack
            # images should be chw, and scalars/vectors should be stacked on dim 0 also
            out[k] = np.concatenate(self.history[k], axis=0)
        return out


class Torchify(gym.Wrapper):
    """Allow the client to use Torch tensors instead of the usual numpy arrays."""

    def to_torch(self, x):
        return map_structure(lambda x: torch.from_numpy(x) if isinstance(x, np.ndarray) else x, x)

    def to_numpy(self, x):
        return map_structure(lambda x: x.numpy() if isinstance(x, torch.Tensor) else x, x)

    def step(self, action):
        action = self.to_numpy(action)
        obs, reward, done, info = self.env.step(action)
        return self.to_torch(obs), float(reward), bool(done), self.to_torch(info)

    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        obs = self.to_torch(obs)
        return obs


class RecordEpisodeStatistics(gym.Wrapper):
    """This wrapper will keep track of cumulative rewards and episode lengths.

    At the end of an episode, the statistics of the episode will be added to ``info``

    Frame skipping is currently **NOT** taken into account.

    I use this instead of gym.RecordEpisodeStatistics because that has a nested info key that I didn't
    want to deal with supporting.
    And I use this wrapper instead of building it into the worker, because we often want to record the reward
    before any reward scaling is applied.
    """

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.episode_return: float = 0
        self.episode_length: int = 0

    def reset(self, **kwargs):
        observations = super().reset(**kwargs)
        self.episode_return = 0
        self.episode_length = 0
        return observations

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.episode_return += reward
        self.episode_length += 1
        info["cumulative_episode_return"] = self.episode_return
        info["cumulative_episode_length"] = self.episode_length
        return obs, reward, done, info



import itertools
from typing import Any
from typing import Optional
from typing import Tuple

import gym
import numpy as np
import torch
import wandb
from einops import rearrange
from einops import repeat
from matplotlib import pyplot as plt
from nptyping import Float32
from nptyping import NDArray
from nptyping import Shape
from skimage.metrics import structural_similarity as ssim_metric
from torch import Tensor
from torch import nn
from torch.distributions import Bernoulli
from torch.distributions import Independent
from torch.distributions import Normal
from tree import flatten
from tree import map_structure

from avalon.agent.common import wandb_lib
from avalon.agent.common.action_model import DictActionDist
from avalon.agent.common.action_model import DictActionHead
from avalon.agent.common.action_model import StraightThroughOneHotCategorical
from avalon.agent.common.action_model import visualize_action_dists
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.types import LatentBatch
from avalon.agent.common.types import ObservationBatch
from avalon.agent.common.types import StateActionBatch
from avalon.agent.common.util import explained_variance
from avalon.agent.common.util import get_avalon_model_seed
from avalon.agent.dreamer.models import RSSM
from avalon.agent.dreamer.models import DenseDecoder
from avalon.agent.dreamer.models import HybridDecoder
from avalon.agent.dreamer.models import HybridEncoder
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.dreamer.tools import lambda_return
from avalon.agent.dreamer.tools import pack_list_of_dicts


class Dreamer(Algorithm[DreamerParams]):
    def __init__(self, params: DreamerParams, obs_space: gym.spaces.Dict, action_space: gym.spaces.Dict) -> None:
        super().__init__(params, obs_space, action_space)

        self._encode = HybridEncoder(obs_space, mlp_hidden_dim=self.params.encoder_mlp_hidden_size)
        feature_size = self.params.deter_size + self.params.stoch_size
        self._decode = HybridDecoder(obs_space, feature_size)
        self._reward = DenseDecoder((), 4, feature_size, self.params.decoder_mlp_hidden_size, dist="normal")
        self._value_current = DenseDecoder((), 4, feature_size, self.params.decoder_mlp_hidden_size, dist="normal")
        self._value_lagged = DenseDecoder((), 4, feature_size, self.params.decoder_mlp_hidden_size, dist="normal")
        self._value_lagged.load_state_dict(self._value_current.state_dict())

        action_head = DictActionHead(action_space, params)
        self._actor = DenseDecoder(
            (action_head.num_inputs,),
            4,
            feature_size,
            self.params.decoder_mlp_hidden_size,
            dist="action_head",
            action_head=action_head,
        )
        self._dynamics = RSSM(
            actdim=action_head.num_outputs,
            embed_size=self._encode.out_dim,
            stoch=self.params.stoch_size,
            deter=self.params.deter_size,
            hidden=self.params.rssm_hidden_size,
        )
        if self.params.pcont:
            # Will use a Bernoulli output dist
            self._pcont = DenseDecoder((), 4, feature_size, self.params.decoder_mlp_hidden_size, dist="binary")

        model_modules = [self._encode, self._dynamics, self._decode, self._reward]
        if self.params.pcont:
            model_modules.append(self._pcont)
        self._model_params = list(itertools.chain(*[list(module.parameters()) for module in model_modules]))
        self._model_opt = torch.optim.Adam(
            self._model_params,
            self.params.model_lr,
            eps=self.params.adam_eps,
            weight_decay=self.params.adam_weight_decay,
        )
        self._value_opt = torch.optim.Adam(
            self._value_current.parameters(),
            self.params.value_lr,
            eps=self.params.adam_eps,
            weight_decay=self.params.adam_weight_decay,
        )
        self._actor_opt = torch.optim.Adam(
            self._actor.parameters(),
            self.params.actor_lr,
            eps=self.params.adam_eps,
            weight_decay=self.params.adam_weight_decay,
        )

        # self.first_step = False
        # if not self.first_step:
        #     logger.info(f"encoder has {sum(p.numel() for p in self._encode.parameters() if p.requires_grad)} params")
        #     logger.info(f"decoder has {sum(p.numel() for p in self._decode.parameters() if p.requires_grad)} params")
        #     logger.info(f"reward has {sum(p.numel() for p in self._reward.parameters() if p.requires_grad)} params")
        #     logger.info(f"dynamics has {sum(p.numel() for p in self._dynamics.parameters() if p.requires_grad)} params")
        #     logger.info(f"value has {sum(p.numel() for p in self._value_current.parameters() if p.requires_grad)} params")
        #     logger.info(f"actor has {sum(p.numel() for p in self._actor.parameters() if p.requires_grad)} params")
        #     # logger.info([(k, p.shape) for k, p in self._actor.named_parameters() if p.requires_grad])
        #     self.first_step = True

        # This is a tuple of (latent, action)
        self.last_rollout_state: Optional[StateActionBatch] = None

    def rollout_step(
        self,
        next_obs: ObservationBatch,
        dones: Tensor,
        indices_to_run: Tensor,
        exploration_mode: str,
    ) -> Tuple[ActionBatch, dict]:
        # Only used in the GamePlayer
        # dones should be "did this env give a done signal after the last step".
        # In other words, obs should follow done. o_t, done_{t-1}

        # This computes the policy given an observation
        # The inputs should all be tensors/structures of tensors on GPU

        if self.last_rollout_state is None:
            device = next_obs[list(next_obs.keys())[0]].device
            batch_size = len(dones)
            prev_latent = self._dynamics.initial(batch_size, device)
            # Using random actions as an initial action probably isn't ideal, but zero isn't a valid 1-hot action..
            # so that seemed worse.
            if (seed := get_avalon_model_seed()) is not None:
                self.action_space.seed(seed)
            prev_action = self.action_space.sample()
            prev_action = map_structure(lambda x: torch.tensor(x, device=device), prev_action)
            prev_action = map_structure(lambda x: repeat(x, "... -> b ...", b=batch_size), prev_action)
            self.last_rollout_state = (prev_latent, prev_action)

        # next_obs = {k: v[indices_to_run] for k, v in next_obs.items()}
        # Check the batch sizes match - that we're not carrying over an old rollout state.
        assert dones.shape[0] == flatten(self.last_rollout_state)[0].shape[0]

        # We want to set done to false for anything that claims to be done but isn't running this step.
        # This will result in no masking for those states.
        # indices_to_run_torch = torch.tensor(indices_to_run, device=dones.device, dtype=torch.bool)
        indices_to_run_torch = indices_to_run.to(device=dones.device)
        dones = dones & indices_to_run_torch
        dones = dones.to(dtype=torch.float32)

        # Mask the state to 0 for any envs that have finished?
        mask = 1 - dones

        # we need this because the action can have a variable number of dims
        def multiply_vector_along_tensor_batch_dim(x: Tensor, vector: Tensor) -> Tensor:
            assert len(vector.shape) == 1
            extra_dims = (1,) * (x.dim() - 1)
            return x * vector.view(-1, *extra_dims)

        self.last_rollout_state = map_structure(
            lambda x: multiply_vector_along_tensor_batch_dim(x, mask) if x is not None else x, self.last_rollout_state
        )
        assert self.last_rollout_state is not None

        # sliced_state = map_structure(lambda x: x[indices_to_run], self.last_rollout_state)
        action, new_state = self.policy(next_obs, self.last_rollout_state, mode=exploration_mode)
        # action, new_state = self.policy(next_obs, sliced_state, mode="explore")

        # Not sure if map_structure works nicely with inplace operations, so we'll do it manually.
        for k1, v1 in enumerate(self.last_rollout_state):
            for k2, v2 in v1.items():
                v2[indices_to_run] = new_state[k1][k2][indices_to_run]

        action = map_structure(lambda x: x.cpu(), action)
        return action, {}

    def reset_state(self) -> None:
        self.last_rollout_state = None

    def policy(
        self, obs: ObservationBatch, prev_state: StateActionBatch, mode: str = "explore"
    ) -> Tuple[ActionBatch, StateActionBatch]:
        """Encode the observation, pass it into the observation model along with the previous state/action
        to generate a new state estimate, and use that to generate a policy.

        state is a tuple of (latent, action)"""

        # Obs is/can be numpy array
        assert prev_state is not None
        prev_latent, prev_action = prev_state

        embed = self._encode(obs)
        latent, _ = self._dynamics.obs_step(prev_latent, prev_action, embed)
        feat = self._dynamics.get_feat(latent)
        if mode == "explore":
            action = self._actor(feat).rsample()
        elif mode == "eval":
            action = self._actor(feat).mode()
        else:
            assert False
        state = (latent, action)
        return action, state

    def _imagine_ahead(self, start: LatentBatch, dones: Tensor) -> dict[str, Any]:
        # In the sequence, at a given index, it's an (action -> state) pair. action comes first.
        # Thus the dummy "0" action at the front.

        start = {k: rearrange(v, "b t n -> (b t) n") for k, v in start.items()}
        start["feat"] = self._dynamics.get_feat(start)
        start["action"] = {k: torch.zeros_like(v) for k, v in self._actor(start["feat"]).rsample().items()}  # type: ignore
        seq = {k: [v] for k, v in start.items()}
        for _ in range(self.params.horizon):
            action = self._actor(seq["feat"][-1].detach()).rsample()
            state = self._dynamics.img_step({k: v[-1] for k, v in seq.items()}, action)
            feat = self._dynamics.get_feat(state)
            for key, value in {**state, "action": action, "feat": feat}.items():
                seq[key].append(value)
        # These now have shape (imag_steps, batch_size * fragment_steps)
        seq_packed = {k: torch.stack(v, 0) for k, v in seq.items() if k != "action"}
        if self.params.pcont:
            disc = self.params.discount * self._pcont(seq_packed["feat"]).probs
            # Override discount prediction for the first step with the true
            # discount factor from the replay buffer.
            dones = rearrange(dones, "b t -> (b t)")
            true_first = 1.0 - dones.float()
            true_first *= self.params.discount
            disc = torch.cat([true_first[None], disc[1:]], 0)
        else:
            disc = self.params.discount * torch.ones(seq_packed["feat"].shape[:-1], device=seq_packed["feat"].device)
        seq_packed["discount"] = disc
        # Shift discount factors because they imply whether the following state
        # will be valid, not whether the current state is valid.
        # TODO: I don't like how the same discount factor is used for value discounting and for this weight.
        # Seems like they have different purposes and should be treated separately.
        seq_packed["weight"] = torch.cumprod(torch.cat([torch.ones_like(disc[:1]), disc[:-1]], 0), 0)
        seq_packed["action"] = pack_list_of_dicts(seq["action"])  # type: ignore
        return seq_packed

    def train_step(self, batch_data: BatchSequenceData, step: int) -> int:
        self.train()
        # Shape of batch_data elements should be (batch_size, timesteps, ...)
        # Images should be (c, h, w), with range [-.5, .5]

        next_obs = batch_data.observation
        actions = batch_data.action
        rewards = batch_data.reward
        is_terminal = batch_data.is_terminal

        batch_size, timesteps = is_terminal.shape

        # Train the encoder and RSSM model. The actor and value models are not used anywhere here.
        embed = self._encode(next_obs)
        # This is where the loop over all timesteps happens
        post, prior = self._dynamics.observe(embed, actions)
        feat = self._dynamics.get_feat(post)
        assert len(feat.shape) == 3
        obs_pred = self._decode(feat)
        # Reinterpret all but the batch dim (no time dim here)
        # Note: log_likelihood of a Normal with constant std can be reinterpreted as a scaled MSE.
        # is a Normal with std 1 appropriate for vectors too? I guess why not, esp since MSE would seem appropriate.
        obs_dists = {k: Independent(Normal(mean, 1), len(mean.shape) - 2) for k, mean in obs_pred.items()}
        obs_likelihoods = {k: v.log_prob(next_obs[k]) for k, v in obs_dists.items()}
        assert all([v.shape == (batch_size, timesteps) for v in obs_likelihoods.values()])
        autoencoder_mask = 1 - is_terminal.int()
        assert autoencoder_mask.shape == (batch_size, timesteps)

        # Note: these logs will have some terms that are masked out in the loss.
        # No great way to log those, i guess we would reshape and slice them out if we were being proper.
        for k, pred in obs_pred.items():
            target = next_obs[k]
            wandb_lib.log_histogram(f"train/observations/{k}_pred", pred, step)
            wandb_lib.log_histogram(f"train/observations/{k}_target", target, step)
            if step % self.params.log_freq_hist == 0:
                # Log EVs of scalar observations
                if len(target.shape) == 3 and target.shape[-1] == 1:
                    ev = explained_variance(pred, target)
                    wandb_lib.log_scalar(f"train/observations/{k}_EV", ev, step, freq=1)

        reward_pred = self._reward(feat)
        # Note we mask out autoencoding of the terminal timestep.
        # This is because that observation actually comes from the start of the next episode (or is masked to grey,
        # as we actually do). We don't want to try to predict this, if we need to know if the episode ended,
        # we have a pcont signal for that.
        # For done but non-terminal (time limit), we should still be given the true next frame, so predicting that is fine.
        likelihoods: dict[str, Tensor] = {}
        # likelihoods["obs"] = sum([(x * autoencoder_mask).mean() for x in obs_likelihoods.values()])
        likelihoods |= {k: (v * autoencoder_mask).mean() for k, v in obs_likelihoods.items()}
        likelihoods["reward"] = reward_pred.log_prob(rewards).mean()
        likelihoods["reward"] *= self.params.reward_loss_scale
        if self.params.pcont:
            # this is "probability of continue" - an estimator of the done signal
            pcont_pred = self._pcont(feat)
            assert type(pcont_pred) == Bernoulli
            # Label smoothing option
            # pcont_target = 1 - 0.9 * is_terminal.float()
            pcont_target = 1 - is_terminal.float()
            loss = -torch.binary_cross_entropy_with_logits(pcont_pred.logits, pcont_target)
            likelihoods["pcont"] = loss.mean()
            likelihoods["pcont"] *= self.params.pcont_loss_scale

            wandb_lib.log_histogram("train/model/pcont/pred", pcont_pred.probs, step)
            wandb_lib.log_histogram("train/model/pcont/target", pcont_target, step)
            wandb_lib.log_scalar("train/model/pcont/loss_mean", -1 * likelihoods["pcont"], step)

        reward_ev = explained_variance(reward_pred.mean, rewards.float())
        wandb_lib.log_scalar("train/reward/ev", reward_ev, step)
        wandb_lib.log_histogram("train/reward/target", rewards.float(), step)
        wandb_lib.log_histogram("train/reward/pred", reward_pred.mean, step)

        prior_dist = self._dynamics.get_dist(prior)
        post_dist = self._dynamics.get_dist(post)

        wandb_lib.log_histogram("train/kl/prior_mean", prior_dist.mean, step)
        wandb_lib.log_histogram("train/kl/prior_std", prior_dist.stddev, step)
        wandb_lib.log_histogram("train/kl/post_mean", prior_dist.mean, step)
        wandb_lib.log_histogram("train/kl/post_std", prior_dist.stddev, step)
        wandb_lib.log_histogram("train/kl/stoch", post["stoch"], step)
        wandb_lib.log_histogram("train/kl/deter", post["deter"], step)
        wandb_lib.log_histogram("train/kl/embed", embed, step)

        # Dreamerv1 approach
        # FWIW, the dreamerv1 approach works differently than the dreamerv2 approach with balance=.5,
        # which doesn't make much sense. Same with kl_loss_scale=2 to account for smaller grads.
        # Emperically, the v1 approach did better at generating high reward EVs in dmc_cartpole_balance
        # div = torch.distributions.kl_divergence(post_dist, prior_dist).mean()
        # div_clipped = torch.maximum(div, torch.tensor(self._c.free_nats))
        # kl_loss = div_clipped
        # wandb_lib.log_scalar("train/model/div", div, step)

        # Dreamerv2 approach.
        kl_loss, kl_value = self._dynamics.kl_loss(
            post, prior, balance=self.params.kl_balance, free=self.params.free_nats
        )
        assert len(kl_loss.shape) == 0
        wandb_lib.log_scalar("train/model/div", kl_value.mean(), step)

        model_loss = self.params.kl_loss_scale * kl_loss - sum(likelihoods.values())
        wandb_lib.log_scalar("train/model/kl_loss_mean", self.params.kl_loss_scale * kl_loss, step)

        self._model_opt.zero_grad(set_to_none=True)
        model_loss.backward()
        model_norm = nn.utils.clip_grad_norm_(self._model_params, self.params.clip_grad_norm)
        self._model_opt.step()

        wandb_lib.log_histogram("train/model/prior_ent", prior_dist.entropy(), step)
        wandb_lib.log_histogram("train/model/post_ent", post_dist.entropy(), step)
        for name, logprob in likelihoods.items():
            wandb_lib.log_scalar(f"train/model/{name}_loss_mean", -logprob.mean(), step)
        wandb_lib.log_scalar("train/model/loss_mean", model_loss.mean(), step)
        wandb_lib.log_scalar("train/model/grad_norm", model_norm.mean(), step)

        ## IMAGINE #############################

        if (self.params.actor_lr > 0 or self.params.value_lr > 0) and step > self.params.freeze_actor_steps:
            # Train the actor model
            seq = self._imagine_ahead({k: v.detach() for k, v in post.items()}, is_terminal)
            # These rewards have not seen any new actions.
            # So they are the rewards received *in* this timestep, ie before another action is taken.
            reward = self._reward(seq["feat"]).mean
            # NOTE: using the "target" value model here. This value computation is used as a target for training the value network too.
            # Wait, how does dynamics loss work if we're not backproping thru the value model?
            # Well, i guess gradients can still pass thru the lagged model even if it's not being trained?
            slow_value = self._value_lagged(seq["feat"]).mean
            disc = seq["discount"]
            weight = seq["weight"].detach()
            # Skipping last time step because it is used for bootstrapping.
            # Value target i corresponds to the state in seq["feat"] i
            value_target = lambda_return(
                reward[:-1], slow_value[:-1], disc[:-1], bootstrap=slow_value[-1], lambda_=self.params.disclam, axis=0
            )
            assert len(value_target) == len(reward) - 1

        ## ACTOR #############################

        # Actions:      0   [a1]  [a2]   a3
        #                  ^  |  ^  |  ^  |
        #                 /   v /   v /   v
        # States:     [z0]->[z1]-> z2 -> z3
        # Targets:     t0   [t1]  [t2]
        # Baselines:  [v0]  [v1]   v2    v3
        # Entropies:        [e1]  [e2]
        # Weights:    [ 1]  [w1]   w2    w3
        # Loss:              l1    l2

        # Reward:      r0   [r1]  [r2]   r3
        # Two states are lost at the end of the trajectory, one for the boostrap
        # value prediction and one because the corresponding action does not lead
        # anywhere anymore. One target is lost at the start of the trajectory
        # because the initial state comes from the replay buffer.

        # Explaining the above:
        # A state is a (s, a) pair. So z1 is the state of the world right before we take action a1.
        # Let's think just about training the policy that predicted a1. a1 was predicted from z0.
        # The reward resulting from taking action a1 is r1. the value of state z1 (containing r1 and future rewards)
        # is v1.
        # The baseline is the value of the 'state' that existed before we decided on action a1. This doesn't exist
        # as a discrete state in this model, but instead we have state z0, which is the same logical state -
        # z0 is the state where we decided on the action before a1, and thus with a perfect model would know the resulting
        # state after stepping the env also.
        # So the baseline is v0.

        # Reinforce works by saying: if things worked out well in the rollout from this state, when action_x was taken,
        # update the policy to make it more likely to take that action.
        # "worked out well" specifically means "worked better than expected" - and "expected" is the value estimate
        # before seeing the action.
        # So we want the value at this state, the policy computed from this state,
        # and the action actually taken after observing that state.

        if self.params.actor_lr > 0 and step > self.params.freeze_actor_steps:
            policy = self._actor(seq["feat"][:-2].detach())
            assert type(policy) == DictActionDist
            if self.params.actor_grad == "dynamics":
                objective = value_target[1:]
            elif self.params.actor_grad == "reinforce":
                # Why do we recompute this here? We compute the same thing above. I guess different gradient flow somehow? but grads don't even go thru this.
                baseline = self._value_lagged(seq["feat"][:-2]).mode
                advantage = (value_target[1:] - baseline).detach()
                _action = {k: v[1:-1].detach() for k, v in seq["action"].items()}
                objective = policy.log_prob(_action) * advantage
            elif self.params.actor_grad == "hybrid":
                # Dynamics works well for continuous, REINFORCE works well for discrete.
                # We can compute REINFORCE just on the discrete actions. Can't split up the dynamics as easily tho
                # without recomputing the imagination rollout with carefully placed detach()es.
                # So here we just compute dynamics loss normally, and add the discrete-only REINFORCE loss.
                # TODO: this is hacky - the categorical should always be wrapped in an independent so this works..
                discrete_policy = DictActionDist(
                    {
                        k: v
                        for k, v in policy.dists.items()
                        if isinstance(v.base_dist, StraightThroughOneHotCategorical)  # type: ignore
                    }
                )
                assert len(discrete_policy.dists) > 0

                baseline = self._value_lagged(seq["feat"][:-2]).mode
                advantage = (value_target[1:] - baseline).detach()
                discrete_action = {
                    k: v[1:-1].detach() for k, v in seq["action"].items() if k in discrete_policy.dists.keys()
                }
                discrete_reinforce_objective = discrete_policy.log_prob(discrete_action) * advantage
                dynamics_objective = value_target[1:]
                assert discrete_reinforce_objective.shape == dynamics_objective.shape
                # Could add mixing weights here
                objective = discrete_reinforce_objective + dynamics_objective
            else:
                assert False

            # Entropy shape should be (batch * fragment_len, imag_len)
            actor_entropy = policy.entropy()
            actor_entropy_loss = actor_entropy * self.params.policy_entropy_scale
            assert objective.shape == actor_entropy_loss.shape
            objective = objective + actor_entropy_loss

            actor_loss = -1 * objective * weight[:-2]
            self._actor_opt.zero_grad(set_to_none=True)
            actor_loss.mean().backward()
            actor_norm = nn.utils.clip_grad_norm_(self._actor.parameters(), self.params.clip_grad_norm)
            self._actor_opt.step()

            wandb_lib.log_histogram("train/actor/loss", actor_loss, step)  # note this includes the entropy penalty
            wandb_lib.log_scalar("train/actor/grad_norm", actor_norm.mean(), step)
            wandb_lib.log_scalar("train/actor/entropy_loss_mean", -1 * actor_entropy_loss.mean(), step)
            wandb_lib.log_histogram("train/actor/entropy", actor_entropy, step)
            wandb_lib.log_histogram("train/actor/steps_imagined", seq["weight"].sum(dim=0), step)
            wandb_lib.log_histogram("train/reward/imagined", reward[:-1], step)

            # Visualize actions
            visualize_action_dists(self.action_space, policy, prefix="imagination_policy")

        ## VALUE #############################
        if self.params.value_lr > 0 and step > self.params.freeze_actor_steps:
            # Train the value model
            # I'm curious why they train the value model in imagination but not in the actual rollouts.
            # Oh, I guess it has to be on-policy and that's only the case in imagination.
            value_pred = self._value_current(seq["feat"].detach()[:-1], raw_feats=True)
            # isn't the log prob of a Normal with std 1 the same as MSE?
            # Not using an Independent is appropriate here since the event is a scalar and the batch is (t, b)
            value_pred = Normal(value_pred, 1)
            value_loss = -1 * (weight[:-1] * value_pred.log_prob(value_target.detach())).mean()

            self._value_opt.zero_grad(set_to_none=True)
            value_loss.backward()
            value_norm = nn.utils.clip_grad_norm_(self._value_current.parameters(), self.params.clip_grad_norm)
            self._value_opt.step()

            # Update the "target" network.
            # TODO: an EMA might be more appropriate.
            if step % self.params.value_target_network_update_freq == 0:
                self._value_lagged.load_state_dict(self._value_current.state_dict())

            value_ev = explained_variance(value_pred.mean * weight[:-1], value_target.float() * weight[:-1])
            wandb_lib.log_scalar("train/value/ev", value_ev, step)
            # TODO: these histograms of things that are soft-masked don't really making sense.
            # Without the mask, we're showing hists containing junk. But if we apply the mask,
            # the hist ends up with wrong values. Ideally we'd make a "weighted hist".
            # In the meantime, just consider they'll be containing junk unless the imag rollouts rarely contain ep ends.
            wandb_lib.log_histogram("train/value/pred_slow", slow_value, step)
            wandb_lib.log_histogram("train/value/pred", value_pred.mean, step)
            wandb_lib.log_histogram("train/value/target", value_target, step)

            wandb_lib.log_scalar("train/value/grad_norm", value_norm.mean(), step)
            wandb_lib.log_scalar("train/value/loss_mean", value_loss.mean(), step)
            wandb_lib.log_histogram("train/value/weight", weight, step)

        # if "rgbd" in next_obs and step % self.params.log_freq_media == 0:
        for k, v in next_obs.items():
            if len(v.shape) == 5 and step % wandb_lib.MEDIA_FREQ == 0:  # b, t, c, h, w
                batch_size = 6
                # Do an imagination rollout. Can we use the imagine_ahead logic instead of replicating here?
                truth = next_obs[k][:batch_size, :, :3] + 0.5
                recon = obs_pred[k][:batch_size, :, :3]
                # we observe the first 5 frames to estimate the state
                sliced_actions = {k: v[:batch_size, :5] for k, v in actions.items()}
                init, _ = self._dynamics.observe(embed[:batch_size, :5], sliced_actions)
                init = {k: v[:, -1] for k, v in init.items()}
                # Then do an imagination rollout from there
                actual_actions = {k: rearrange(v[:batch_size, 5:], "b t ... -> t b ...") for k, v in actions.items()}
                prior = self._dynamics.imagine(actual_actions, init)
                openl = self._decode(self._dynamics.get_feat(prior))[k][:, :, :3]
                # First 5 frames are recon, next are imagination
                model = torch.cat([recon[:, :5] + 0.5, openl + 0.5], 1)
                error = (model - truth + 1) / 2
                comparison = torch.cat([truth, model, error], 3)
                wandb_lib.log_video("video/openl", comparison, step, freq=1, num_images_per_row=batch_size)

                # similarity metrics
                truth_np = truth.detach().cpu().numpy()
                recon_np = recon.detach().cpu().numpy()
                openl_np = openl.detach().cpu().numpy()
                recon_ssim = calculate_framewise_ssim(truth_np + 0.5, recon_np + 0.5)
                openl_ssim = calculate_framewise_ssim(truth_np[:, 5:] + 0.5, openl_np + 0.5)
                recon_ssim_plot = make_ssim_plot(recon_ssim, "Reconstructed observation SSIM by frame across batches")
                openl_ssim_plot = make_ssim_plot(openl_ssim, "Imagination rollout SSIM by frame across batches")
                wandb.log({f"images/recon_ssim": wandb.Image(recon_ssim_plot)})
                wandb.log({f"images/openl_ssim": wandb.Image(openl_ssim_plot)})
        return step + 1


NPObservationBatch = NDArray[Shape["Batch, Frame, Channel, Height, Width"], Float32]
NPFramewiseMetric = NDArray[Shape["Batch, Frame"], Float32]


def calculate_framewise_ssim(ground_truth: NPObservationBatch, predictions: NPObservationBatch) -> NPFramewiseMetric:
    ssim = np.zeros(ground_truth.shape[:2])
    batch_size, fragment_length = ssim.shape
    for batch in range(batch_size):
        for frame in range(fragment_length):
            ssim[batch, frame] = ssim_metric(ground_truth[batch, frame], predictions[batch, frame], channel_axis=0)
    return ssim


def make_ssim_plot(ssim: NPFramewiseMetric, title: str):
    fig, ax = plt.subplots(1)
    plt.plot(ssim.T)
    plt.ylabel("SSIM")
    plt.xlabel("Distance in frames")
    plt.title(title)
    return fig



import math
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np
from loguru import logger
from scipy import stats
from scipy.ndimage import convolve

from avalon.common.errors import SwitchError
from avalon.datagen.godot_base_types import Vector2
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import euclidean_distance
from avalon.datagen.world_creation.indoor.constants import TILE_SIZE
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.constants import CornerType
from avalon.datagen.world_creation.indoor.constants import WallType
from avalon.datagen.world_creation.indoor.wall_footprint import WallFootprint
from avalon.datagen.world_creation.types import BuildingBoolNP
from avalon.datagen.world_creation.types import BuildingFloatNP
from avalon.datagen.world_creation.types import BuildingIntNP
from avalon.datagen.world_creation.types import Point3DNP

Corner = Tuple[Tuple[int, int], Optional[CornerType], bool]


def find_corners(tiles: BuildingIntNP, solid_tile_value: int = 1) -> List[Corner]:
    """
    Iterates through tiles to find all interior/exterior corners of a wall in order.
    Rough algorithm:
    1. Iterate through tiles east-ward until it finds a wall (denoted by `solid_tile_value`)
    2. Trace along the wall starting from the first point, attempting to make a turn
    3. If a turn is possible, a corner is found and direction changes; if not, keep going the same direction
    4. Repeat steps 2-3 until we reach the initial corner again - then all corners have been found.
    """
    length, width = tiles.shape
    current_direction = Azimuth.EAST
    z, x = 0, 0
    done = False
    corners: List[Corner] = []
    while not done:
        if not corners:
            if tiles[z, x] == solid_tile_value:
                corners.append(((z, x), None, True))
            else:
                if x < width - 1:
                    x += 1
                    continue
                elif z < length - 1:
                    x = 0
                    z += 1
                    continue
                else:
                    break  # no corners

        # Always try turning left first
        directions = [Azimuth.NORTH, Azimuth.EAST, Azimuth.SOUTH, Azimuth.WEST] * 2
        current_idx = directions.index(current_direction)
        sorted_dirs = [directions[current_idx - 1], *directions[current_idx : current_idx + 3]]
        for i, new_direction in enumerate(sorted_dirs):
            new_z, new_x = z, x
            if new_direction == Azimuth.NORTH:
                new_z -= 1
            elif new_direction == Azimuth.EAST:
                new_x += 1
            elif new_direction == Azimuth.SOUTH:
                new_z += 1
            elif new_direction == Azimuth.WEST:
                new_x -= 1
            else:
                raise SwitchError(new_direction)

            if new_z < 0 or new_x < 0:
                continue  # negative = out of bounds

            try:
                tile = tiles[new_z, new_x]
            except IndexError:
                continue  # out of bounds

            if tile != solid_tile_value:
                continue  # outside footprint

            last_coordinates = (z, x)
            z, x = new_z, new_x
            if new_direction == current_direction:
                break  # keep going along the border, no corner yet

            is_outside_corner = i != 0
            type_mapping = {
                (Azimuth.NORTH, Azimuth.EAST): (CornerType.NW,),
                (Azimuth.NORTH, Azimuth.WEST): (CornerType.NE,),
                (Azimuth.NORTH, Azimuth.SOUTH): (CornerType.NW, CornerType.NE),
                (Azimuth.EAST, Azimuth.NORTH): (CornerType.SE,),
                (Azimuth.EAST, Azimuth.SOUTH): (CornerType.NE,),
                (Azimuth.EAST, Azimuth.WEST): (CornerType.NE, CornerType.SE),
                (Azimuth.SOUTH, Azimuth.EAST): (CornerType.SW,),
                (Azimuth.SOUTH, Azimuth.WEST): (CornerType.SE,),
                (Azimuth.SOUTH, Azimuth.NORTH): (CornerType.SE, CornerType.SW),
                (Azimuth.WEST, Azimuth.SOUTH): (CornerType.NW,),
                (Azimuth.WEST, Azimuth.NORTH): (CornerType.SW,),
                (Azimuth.WEST, Azimuth.EAST): (CornerType.SW, CornerType.NW),
            }
            new_corners: List[Corner] = []
            for corner_type in type_mapping[(current_direction, new_direction)]:
                new_corners.append((last_coordinates, corner_type, is_outside_corner))
            if last_coordinates == corners[0][0]:
                corners.pop(0)
                corners = new_corners + corners
                done = True
            else:
                corners.extend(new_corners)
                current_direction = new_direction
            break
    return corners


def find_exterior_wall_footprints(
    tiles: BuildingIntNP, no_build_mask: Optional[BuildingBoolNP] = None
) -> List[WallFootprint]:
    walls = []
    for wall_type in WallType:
        wall_kernel = wall_type.convolution_kernel
        out_of_bounds = tiles == False
        wall_or_corner_mask = convolve(tiles.astype(np.int8), wall_kernel, mode="constant") == 0
        wall_mask = ~out_of_bounds & wall_or_corner_mask
        for corner_type in wall_type.corner_types:
            corner_kernel = corner_type.convolution_kernel
            corner_mask = convolve(tiles.astype(np.int8), corner_kernel, mode="constant") == 0
            wall_mask &= ~corner_mask

        # Hallways and story links (ramps) intersecting with exterior walls are entrances/exits; if mask is passed,
        # we exclude them; e.g. to avoid building windows there.
        if no_build_mask is not None:
            wall_mask &= ~no_build_mask

        if wall_type.is_vertical:
            wall_mask = wall_mask.T
        wall_tiles = list(zip(*np.where(wall_mask)))

        wall = None
        for current_tile, next_tile in zip(wall_tiles[:-1], wall_tiles[1:]):
            if not wall:
                z, x = current_tile
                wall = [(z, x) if wall_type.is_vertical else (x, z)]

            end_tile = None
            if current_tile[1] + 1 != next_tile[1]:
                end_tile = current_tile
            elif wall is not None and next_tile == wall_tiles[-1]:
                end_tile = next_tile

            if end_tile:
                z, x = end_tile
                z += 1
                x += 1
                wall.append((z, x) if wall_type.is_vertical else (x, z))
                walls.append(WallFootprint(wall[0], wall[1], wall_type, wall_type.is_vertical))
                wall = None
    return walls


def visualize_tiles(tiles: BuildingIntNP) -> None:
    char_by_int = {i + 1: chr(ord("A") + i) for i in range(0, 52)}
    char_by_int[0] = "."
    for line in tiles:
        chars = [char_by_int[-x if x < 0 else x] for x in line]
        logger.debug("".join(chars))


def draw_line_in_grid(
    points: Tuple[BuildingTile, ...],
    grid: Union[BuildingIntNP, BuildingFloatNP],
    line_tile_value: Union[int, float],
    drawable_grid_value: Optional[Union[int, float]] = 0,
    include_ends: bool = True,
) -> None:
    for i, (from_point, to_point) in enumerate(zip(points[:-1], points[1:])):
        is_first_connection = i == 0
        is_last_connection = i == len(points) - 2
        is_connection_vertical = from_point.x == to_point.x
        if is_connection_vertical:
            axis = Axis.Z.value
        else:
            axis = Axis.X.value

        # We call max(..., 0) here for negative coordinates (exterior points) to work correctly
        start_coord = max(min(getattr(from_point, axis), getattr(to_point, axis)), 0)
        end_coord = max(getattr(from_point, axis), getattr(to_point, axis)) + 1
        if not include_ends:
            if is_first_connection and start_coord != 0:
                start_coord += 1
            if is_last_connection:
                end_coord -= 1

        if is_connection_vertical:
            local_grid = grid[start_coord:end_coord, from_point.x]
        else:
            local_grid = grid[from_point.z, start_coord:end_coord]

        if drawable_grid_value is not None:
            local_grid[local_grid == drawable_grid_value] = line_tile_value
            if include_ends:
                # todo: this is sorta leaky
                local_grid[0] = line_tile_value
                local_grid[-1] = line_tile_value
        else:
            local_grid[True] = line_tile_value


def tile_centroid(tile_position: Tuple[int, int], tile_size: float = 1) -> Vector2:
    half_size = tile_size / 2
    return Vector2(tile_position[0] + half_size, tile_position[1] + half_size)


def decide_tiles_by_distance(
    free_tiles: List[Tuple[int, int]],
    target_tile: Tuple[int, int],
    difficulty: float,
    rand: np.random.Generator,
    tile_count: int = 1,
) -> List[Tuple[int, int]]:
    target_tile_position = BuildingTile(*target_tile)
    tile_distances = [
        euclidean_distance(BuildingTile(*tile_position), target_tile_position) for tile_position in free_tiles
    ]
    desired_distance = difficulty * max(tile_distances)
    std = np.std(tile_distances)
    std = std if std != 0 else 1  # std=0 yields invalid distribution
    distance_distribution = stats.norm(desired_distance, math.sqrt(std))
    weights = np.array([distance_distribution.pdf(distance) for distance in tile_distances])
    weights /= weights.sum()
    return [tuple(tile) for tile in rand.choice(free_tiles, tile_count, p=weights)]  # type: ignore


def get_neighbor_tiles(
    tiles: np.ndarray, point: Tuple[int, int], only_if_equals: Optional[Any] = None
) -> Tuple[Tuple[int, int], ...]:
    z, x = point
    length, width = tiles.shape
    neighbor_tiles = []
    if z + 1 < length:
        neighbor_tiles.append((z + 1, x))
    if x + 1 < width:
        neighbor_tiles.append((z, x + 1))
    if z - 1 >= 0:
        neighbor_tiles.append((z - 1, x))
    if x - 1 >= 0:
        neighbor_tiles.append((z, x - 1))
    return tuple(
        [tile for tile in neighbor_tiles if only_if_equals is None or tiles[tile[0], tile[1]] == only_if_equals]
    )


def is_point_in_tile(point: Point3DNP, tile: BuildingTile) -> bool:
    x, _y, z = point
    return cast(bool, tile.x < x < tile.x + TILE_SIZE and tile.z < z < tile.z + TILE_SIZE)



import json
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Final
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypedDict
from typing import Union

import numpy as np
from einops import rearrange
from loguru import logger
from numpy.typing import NDArray

from avalon.agent.godot.godot_gym import create_base_benchmark_config
from avalon.common.utils import wait_until_true
from avalon.common.visual_utils import visualize_arraylike_as_video
from avalon.datagen.godot_env.actions import AttrsAction
from avalon.datagen.godot_env.actions import MouseKeyboardAction
from avalon.datagen.godot_env.actions import VRAction
from avalon.datagen.godot_env.goals import NullGoalEvaluator
from avalon.datagen.godot_env.godot_env import GodotEnv
from avalon.datagen.godot_env.observations import AttrsObservation
from avalon.datagen.godot_env.observations import AvalonObservation
from avalon.datagen.godot_env.replay import get_action_type_from_config as _get_action_type_from_config
from avalon.datagen.godot_generated_types import AgentPlayerSpec
from avalon.datagen.godot_generated_types import AvalonSimSpec
from avalon.datagen.godot_generated_types import MouseKeyboardAgentPlayerSpec
from avalon.datagen.godot_generated_types import VRAgentPlayerSpec
from avalon.datagen.godot_utils import S3_AVALON_ERROR_BUCKET

_ActOrStepObservationSequence = Union[Sequence[AvalonObservation], Sequence[Dict[str, Any]]]

get_action_type_from_config: Final = _get_action_type_from_config


def create_mouse_keyboard_benchmark_config() -> AvalonSimSpec:
    with create_base_benchmark_config().mutable_clone() as config:
        assert isinstance(config.player, AgentPlayerSpec)
        config.player = MouseKeyboardAgentPlayerSpec.from_dict(config.player.to_dict())
        return config


def create_vr_benchmark_config() -> AvalonSimSpec:
    with create_base_benchmark_config().mutable_clone() as config:
        assert isinstance(config.player, AgentPlayerSpec)
        config.player = VRAgentPlayerSpec.from_dict(config.player.to_dict())
        return config


def get_null_mouse_keyboard_action() -> AttrsAction:
    return MouseKeyboardAction(
        head_x=0.0,
        head_z=0.0,
        head_pitch=0.0,
        head_yaw=0.0,
        is_left_hand_grasping=0.0,
        is_right_hand_grasping=0.0,
        is_left_hand_throwing=0.0,
        is_right_hand_throwing=0.0,
        is_jumping=0.0,
        is_eating=0.0,
        is_crouching=0.0,
    )


def get_null_vr_action() -> VRAction:
    return VRAction(
        head_x=0.0,
        head_y=0.0,
        head_z=0.0,
        head_pitch=0.0,
        head_yaw=0.0,
        head_roll=0.0,
        left_hand_x=0.0,
        left_hand_y=0.0,
        left_hand_z=0.0,
        left_hand_pitch=0.0,
        left_hand_yaw=0.0,
        left_hand_roll=0.0,
        is_left_hand_grasping=0.0,
        right_hand_x=0.0,
        right_hand_y=0.0,
        right_hand_z=0.0,
        right_hand_pitch=0.0,
        right_hand_yaw=0.0,
        right_hand_roll=0.0,
        is_right_hand_grasping=0.0,
        is_jumping=0.0,
    )


def create_env(
    config: AvalonSimSpec,
    action_type: Type[AttrsAction],
    is_logging_artifacts_on_error_to_s3: bool = False,
    observation_type: Type[AttrsObservation] = AvalonObservation,
) -> GodotEnv:
    return GodotEnv(
        config=config,
        observation_type=observation_type,
        action_type=action_type,
        goal_evaluator=NullGoalEvaluator(),
        is_logging_artifacts_on_error_to_s3=is_logging_artifacts_on_error_to_s3,
        s3_bucket_name=S3_AVALON_ERROR_BUCKET,
    )


def rgbd_to_video_array(rgbd_data: Iterable[NDArray]) -> NDArray:
    return np.stack(
        [
            rearrange(
                np.flipud(rgbd[:, :, :3]),
                "h w c -> c h w",
            )
            / 255.0
            for rgbd in rgbd_data
        ]
    )


def observation_video_array(data: _ActOrStepObservationSequence) -> NDArray:
    if len(data) == 0:
        return rgbd_to_video_array([])

    if isinstance(data[0], dict):
        assert "rgbd" in data[0], "observation dict does not contain rgbd data"
        return rgbd_to_video_array(x["rgbd"] for x in data)  # type: ignore[index]

    return rgbd_to_video_array(x.rgbd for x in data)  # type: ignore[union-attr]


def display_video(
    data: _ActOrStepObservationSequence,
    size: Optional[Tuple[int, int]] = None,
    fps: int = 25,
    file_format: str = "webm",
) -> None:
    if size is None:
        size = (512, 512)
    video_array = observation_video_array(data)
    visualize_arraylike_as_video(
        video_array,
        normalize=False,
        size=size,
        video_format=file_format,
        fps=fps,
    )


def visualize_worlds_in_folder(world_paths: Iterable[Path], resolution: int = 1024, num_frames: int = 20):
    episode_seed = 0
    config = create_vr_benchmark_config()

    with config.mutable_clone() as config:
        config.recording_options.resolution_x = resolution
        config.recording_options.resolution_y = resolution
    action_type = VRAction
    env = create_env(config, action_type)

    all_observations = []
    # if we want to take a few actions
    # null_action = get_null_vr_action()
    worlds_to_sort = []
    for world_path in world_paths:
        task, seed_str, difficulty_str = world_path.name.split("__")
        difficulty = float(difficulty_str.replace("_", "."))
        seed = int(seed_str)
        worlds_to_sort.append((task, difficulty, seed, world_path))

    for (task, difficulty, seed, world_path) in sorted(worlds_to_sort):
        logger.debug(f"Loading {world_path}")
        world_file = world_path / "main.tscn"
        observations = []
        observations.append(
            env.reset_nicely_with_specific_world(
                episode_seed=episode_seed,
                world_path=str(world_file),
            )
        )
        for i in range(num_frames):
            null_action = get_null_vr_action()
            obs, _ = env.act(null_action)
            observations.append(obs)

        all_observations.append(observations)

    return all_observations


DebugItemLog = TypedDict(
    "DebugItemLog",
    {
        "name": str,
        "id": str,
        # TODO not properly overridden in children
        "class": str,
        "script": str,
        "position": str,
        "velocity": str,
        "rotation": str,
    },
)


class DebugAnimalLog(DebugItemLog):
    is_frozen: bool
    is_mobile: bool
    is_alive: bool
    hit_points: int
    current_domain: str
    behavior: str
    forward_impediment: str


# TODO make serializable version with "unknown_attributes" attr to avoid fragile debug logs
class DebugLogLine(TypedDict):
    time: int
    episode: int
    items: List[Union[DebugItemLog, DebugAnimalLog]]


def _output_folder_for(env: GodotEnv, episode_seed: int) -> Path:
    dir_root = env.config.get_dir_root()
    return Path(dir_root) / f"{episode_seed:06d}"


def get_debug_json_logs(env: GodotEnv, episode_seed: Optional[int] = None) -> List[DebugLogLine]:
    assert (
        env.config.recording_options.is_debugging_output_requested
    ), "env.config.recording_options.is_debugging_output_requested is False, would not have produced a debug log"
    if episode_seed is None:
        episode_seed = env._latest_episode_seed
    is_episode_ongoing = not env.process.is_closed and episode_seed == env._latest_episode_seed
    debug_path = _output_folder_for(env, episode_seed) / "debug.json"
    return _read_debug_json_log(debug_path, is_episode_ongoing=is_episode_ongoing)


def _read_debug_json_log(debug_json_path: Path, is_episode_ongoing: bool) -> List[DebugLogLine]:
    debug_output: List[DebugLogLine] = []
    failure = None
    if is_episode_ongoing:
        wait_until_true(debug_json_path.exists, sleep_inc=0.1)
    with open(debug_json_path, "r") as line_delimited_debug_json:
        for line, frame_dict in enumerate(line_delimited_debug_json):
            try:
                debug_output.append(json.loads(frame_dict))
            except json.JSONDecodeError as e:
                failure = (line, e, frame_dict)
                break

        if failure is not None:
            line, err, frame_dict = failure
            next_line = next(line_delimited_debug_json, None)
            is_unparseable_mid_file = next_line is not None
            if is_unparseable_mid_file:
                raise ValueError(
                    f"Failed to decode debug log line {line} '{frame_dict}'."
                    f"This is likely because the episode number is being tinkered with, "
                    f"and the latest debug output is shorter than a previous one. "
                    f"Line: '{frame_dict}'"
                ) from err
            elif not is_episode_ongoing:
                raise ValueError(
                    f"Failed to decode final debug log line {line}  '{frame_dict}'. "
                    f"The epsiode may still be in progress and mid-write."
                ) from err

        return debug_output


class ActionSmoother:
    def __init__(self, decay: float = 0.999) -> None:
        self.state: dict[str, float] = {}
        self.decay = decay

    def update(self, action: dict[str, Any]):
        for k, v in action.items():
            if v.dtype == np.float32:
                if k not in self.state:
                    self.state[k] = v * 0
                else:
                    self.state[k] = v * (1 - self.decay) + self.state[k] * self.decay
            else:
                self.state[k] = v
        return self.state



from typing import Optional

import attr
import numpy as np

from avalon.common.utils import to_immutable_array
from avalon.datagen.world_creation.entities.item import InstancedDynamicItem
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Tool(InstancedDynamicItem):
    position: Point3DNP = attr.ib(
        default=np.array([0.0, 0.0, 0.0]),
        converter=to_immutable_array,
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
    )
    solution_mask: Optional[MapBoolNP] = attr.ib(default=None, eq=False)



""""
Client interface for EvalAI.
Copied mostly without modification from:
https://github.com/Cloud-CV/EvalAI-Starters/blob/master/remote_challenge_evaluation/evaluation_script_starter.py
"""
import requests
from loguru import logger

URLS = {
    "get_message_from_sqs_queue": "/api/jobs/challenge/queues/{}/",
    "get_submission_by_pk": "/api/jobs/submission/{}",
    "delete_message_from_sqs_queue": "/api/jobs/queues/{}/",
    "update_submission": "/api/jobs/challenge/{}/update_submission/",
}


class EvalAI_Interface:
    def __init__(self, AUTH_TOKEN, EVALAI_API_SERVER, QUEUE_NAME, CHALLENGE_PK) -> None:  # type: ignore
        """Class to initiate call to EvalAI backend

        Arguments:
            AUTH_TOKEN {[string]} -- The authentication token corresponding to EvalAI
            EVALAI_API_SERVER {[string]} -- It should be set to https://eval.ai # For production server
            QUEUE_NAME {[string]} -- Unique queue name corresponding to every challenge
            CHALLENGE_PK {[integer]} -- Primary key corresponding to a challenge
        """

        self.AUTH_TOKEN = AUTH_TOKEN
        self.EVALAI_API_SERVER = EVALAI_API_SERVER
        self.QUEUE_NAME = QUEUE_NAME
        self.CHALLENGE_PK = CHALLENGE_PK

    def get_request_headers(self):  # type: ignore
        """Function to get the header of the EvalAI request in proper format

        Returns:
            [dict]: Authorization header
        """
        headers = {"Authorization": "Bearer {}".format(self.AUTH_TOKEN)}
        return headers

    def make_request(self, url, method, data=None):  # type: ignore
        """Function to make request to EvalAI interface

        Args:
            url ([str]): URL of the request
            method ([str]): Method of the request
            data ([dict], optional): Data of the request. Defaults to None.

        Returns:
            [JSON]: JSON response data
        """
        headers = self.get_request_headers()
        try:
            response = requests.request(method=method, url=url, headers=headers, data=data)
            response.raise_for_status()
        except requests.exceptions.RequestException:
            logger.info("The server isn't able establish connection with EvalAI")
            raise
        return response.json()

    def return_url_per_environment(self, url):  # type: ignore
        """Function to get the URL for API

        Args:
            url ([str]): API endpoint url to which the request is to be made

        Returns:
            [str]: API endpoint url with EvalAI base url attached
        """
        base_url = "{0}".format(self.EVALAI_API_SERVER)
        url = "{0}{1}".format(base_url, url)
        return url

    def get_message_from_sqs_queue(self):  # type: ignore
        """Function to get the message from SQS Queue

        Docs: https://eval.ai/api/docs/#operation/get_submission_message_from_queue

        Returns:
            [JSON]: JSON response data
        """
        url = URLS.get("get_message_from_sqs_queue").format(self.QUEUE_NAME)
        url = self.return_url_per_environment(url)
        response = self.make_request(url, "GET")
        return response

    def delete_message_from_sqs_queue(self, receipt_handle):  # type: ignore
        """Function to delete the submission message from the queue

        Docs: https://eval.ai/api/docs/#operation/delete_submission_message_from_queue

        Args:
            receipt_handle ([str]): Receipt handle of the message to be deleted

        Returns:
            [JSON]: JSON response data
        """
        url = URLS.get("delete_message_from_sqs_queue").format(self.QUEUE_NAME)
        url = self.return_url_per_environment(url)
        data = {"receipt_handle": receipt_handle}
        response = self.make_request(url, "POST", data)
        return response

    def update_submission_data(self, data):  # type: ignore
        """Function to update the submission data on EvalAI

        Docs: https://eval.ai/api/docs/#operation/update_submission

        Args:
            data ([dict]): Data to be updated

        Returns:
            [JSON]: JSON response data
        """
        url = URLS.get("update_submission").format(self.CHALLENGE_PK)
        url = self.return_url_per_environment(url)
        response = self.make_request(url, "PUT", data=data)
        return response

    def update_submission_status(self, data):  # type: ignore
        """

        Docs: https://eval.ai/api/docs/#operation/update_submission

        Args:
            data ([dict]): Data to be updated

        Returns:
            [JSON]: JSON response data
        """
        url = URLS.get("update_submission").format(self.CHALLENGE_PK)
        url = self.return_url_per_environment(url)
        response = self.make_request(url, "PATCH", data=data)
        return response

    def get_submission_by_pk(self, submission_pk):  # type: ignore
        url = URLS.get("get_submission_by_pk").format(submission_pk)
        url = self.return_url_per_environment(url)
        response = self.make_request(url, "GET")
        return response



from __future__ import annotations

import attr
from torch import Tensor

from avalon.agent.common.types import BatchData
from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.types import SequenceData
from avalon.agent.common.types import StepData


@attr.s(auto_attribs=True, frozen=True)
class PPOBatchSequenceData(BatchSequenceData):
    value: Tensor
    policy_prob: Tensor
    policy_entropy: Tensor


@attr.s(auto_attribs=True, frozen=True)
class PPOSequenceData(SequenceData):
    value: Tensor
    policy_prob: Tensor
    policy_entropy: Tensor


@attr.s(auto_attribs=True, frozen=True)
class PPOBatchSequenceDataWithGAE(PPOBatchSequenceData):
    # These get added after GAE computation
    advantage: Tensor
    reward_to_go: Tensor


PPOBatchDataWithGAE = PPOBatchSequenceDataWithGAE  # these have shape (batch, ...) instead of (batch, timesteps, ...)


@attr.s(auto_attribs=True, frozen=True)
class PPOBatchData(BatchData):
    value: Tensor
    policy_prob: Tensor
    policy_entropy: Tensor


@attr.s(auto_attribs=True, frozen=True)
class PPOStepData(StepData):
    batch_type = PPOBatchData
    sequence_type = PPOSequenceData  # type: ignore
    batch_sequence_type = PPOBatchSequenceData  # type: ignore
    value: float
    policy_prob: float
    policy_entropy: float



"""this implementation is from https://github.com/toshas/torch_truncnorm/blob/main/TruncatedNormal.py"""
import math
from numbers import Number
from typing import Optional
from typing import Union

import torch
from torch import Tensor
from torch.distributions import Distribution
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all

CONST_SQRT_2 = math.sqrt(2)
CONST_INV_SQRT_2PI = 1 / math.sqrt(2 * math.pi)
CONST_INV_SQRT_2 = 1 / math.sqrt(2)
CONST_LOG_INV_SQRT_2PI = math.log(CONST_INV_SQRT_2PI)
CONST_LOG_SQRT_2PI_E = 0.5 * math.log(2 * math.pi * math.e)

TensorOrNumber = Union[Tensor, float, int]


class TruncatedStandardNormal(Distribution):
    """
    Truncated Standard Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    arg_constraints = {
        "a": constraints.real,
        "b": constraints.real,
    }
    has_rsample = True

    def __init__(self, a: TensorOrNumber, b: TensorOrNumber, validate_args: Optional[bool] = None) -> None:
        self.a: Tensor
        self.b: Tensor
        self.a, self.b = broadcast_all(a, b)
        if isinstance(a, Number) and isinstance(b, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.a.size()
        super(TruncatedStandardNormal, self).__init__(batch_shape, validate_args=validate_args)
        if self.a.dtype != self.b.dtype:
            raise ValueError("Truncation bounds types are different")
        if any((self.a >= self.b).view(-1).tolist()):
            raise ValueError("Incorrect truncation range")
        eps = torch.finfo(self.a.dtype).eps
        self._dtype_min_gt_0 = eps
        self._dtype_max_lt_1 = 1 - eps
        self._little_phi_a = self._little_phi(self.a)
        self._little_phi_b = self._little_phi(self.b)
        # This explodes (becomes 1, which explodes stuff downstream) if a > ~4
        self._big_phi_a = self._big_phi(self.a)
        self._big_phi_b = self._big_phi(self.b)
        self._Z: Tensor = (self._big_phi_b - self._big_phi_a).clamp_min(eps)
        self._log_Z = self._Z.log()
        little_phi_coeff_a = torch.nan_to_num(self.a, nan=math.nan)
        little_phi_coeff_b = torch.nan_to_num(self.b, nan=math.nan)
        self._lpbb_m_lpaa_d_Z = (
            self._little_phi_b * little_phi_coeff_b - self._little_phi_a * little_phi_coeff_a
        ) / self._Z
        self._mean: Tensor = -(self._little_phi_b - self._little_phi_a) / self._Z
        self._variance: Tensor = 1 - self._lpbb_m_lpaa_d_Z - ((self._little_phi_b - self._little_phi_a) / self._Z) ** 2
        self._entropy: Tensor = CONST_LOG_SQRT_2PI_E + self._log_Z - 0.5 * self._lpbb_m_lpaa_d_Z
        self._mode: Tensor = torch.clip(torch.zeros_like(self.a), self.a, self.b)

    @constraints.dependent_property
    def support(self):
        return constraints.interval(self.a, self.b)

    @property
    def mean(self) -> Tensor:
        return self._mean

    def mode(self) -> Tensor:
        return self._mode

    @property
    def variance(self) -> Tensor:
        return self._variance

    @property
    def stddev(self):
        return self.variance.pow(0.5)

    # @property
    def entropy(self) -> Tensor:
        return self._entropy

    @property
    def auc(self) -> Tensor:
        return self._Z

    @staticmethod
    def _little_phi(x: Tensor) -> Tensor:
        return (-(x**2) * 0.5).exp() * CONST_INV_SQRT_2PI  # type:  ignore

    @staticmethod
    def _big_phi(x: Tensor) -> Tensor:
        return 0.5 * (1 + (x * CONST_INV_SQRT_2).erf())

    @staticmethod
    def _inv_big_phi(x: Tensor) -> Tensor:
        return CONST_SQRT_2 * (2 * x - 1).erfinv()

    def cdf(self, value: Tensor) -> Tensor:
        if self._validate_args:
            self._validate_sample(value)
        return ((self._big_phi(value) - self._big_phi_a) / self._Z).clamp(0, 1)

    def icdf(self, value: Tensor) -> Tensor:
        return self._inv_big_phi(self._big_phi_a + value * self._Z)

    def log_prob(self, value: Tensor) -> Tensor:
        if self._validate_args:
            self._validate_sample(value)
        return CONST_LOG_INV_SQRT_2PI - self._log_Z - (value**2) * 0.5  # type: ignore

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> Tensor:
        shape = self._extended_shape(sample_shape)
        p = torch.empty(shape, device=self.a.device).uniform_(self._dtype_min_gt_0, self._dtype_max_lt_1)
        sample = self.icdf(p)
        assert torch.all(torch.isfinite(sample))
        return sample


class TruncatedNormal(TruncatedStandardNormal):
    """
    Truncated Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    has_rsample = True

    def __init__(
        self,
        loc: TensorOrNumber,
        scale: TensorOrNumber,
        low: TensorOrNumber,
        high: TensorOrNumber,
        validate_args: Optional[bool] = None,
    ) -> None:
        self.loc: Tensor
        self.scale: Tensor
        self.low: Tensor
        self.high: Tensor
        self.loc, self.scale, self.low, self.high = broadcast_all(loc, scale, low, high)
        a = (self.low - self.loc) / self.scale
        b = (self.high - self.loc) / self.scale
        assert torch.all((b - a) > 0)
        super(TruncatedNormal, self).__init__(a, b, validate_args=validate_args)
        self._log_scale = self.scale.log()
        self._mean = self._mean * self.scale + self.loc
        self._variance = self._variance * self.scale**2
        self._entropy += self._log_scale
        self._mode = torch.clip(self.loc, self.low, self.high)

    def _to_std_rv(self, value: Tensor) -> Tensor:
        return (value - self.loc) / self.scale

    def _from_std_rv(self, value: Tensor) -> Tensor:
        return value * self.scale + self.loc

    def cdf(self, value: Tensor) -> Tensor:
        return super(TruncatedNormal, self).cdf(self._to_std_rv(value))

    def icdf(self, value: Tensor) -> Tensor:
        return self._from_std_rv(super(TruncatedNormal, self).icdf(value))

    def log_prob(self, value: Tensor) -> Tensor:
        return super(TruncatedNormal, self).log_prob(self._to_std_rv(value)) - self._log_scale


class ClippedTruncatedNormal(TruncatedNormal):
    """Matches the Dreamerv2 impl, although I don't get what this is for/why it's necessary."""

    def __init__(
        self, loc: TensorOrNumber, scale: TensorOrNumber, low: TensorOrNumber, high: TensorOrNumber, clip: float = 1e-6
    ) -> None:
        super().__init__(loc, scale, low, high)
        self.lower_clip: Tensor = low + clip  # type: ignore
        self.upper_clip: Tensor = high - clip  # type: ignore

    def rsample(self, *args, **kwargs) -> Tensor:  # type: ignore
        event = super().rsample(*args, **kwargs)
        # TODO: what's this clipping for? applied after drawing samples, just slight pulls in samples
        # that are too close to the bounds. only applied in forward pass. but why?
        # would make sense if the actions are being arctanh'ed for some reason and the edges have really steep slope.
        # but this doesn't seem common/logical?
        # or maybe this is just to ensure we don't pick a value exactly equal to the limit (is that even possible?)
        clipped = torch.clip(event, self.lower_clip, self.upper_clip)
        # Apply clipping in forward but not backward pass.
        event = event - event.detach() + clipped.detach()
        return event

    def sample(self, *args, **kwargs) -> Tensor:  # type: ignore
        return self.rsample(*args, **kwargs).detach()
